import re

from .._text import Text
from ..._validation_error import ValidationError


class Alphanumeric(Text):
    """A setting for entering text values limited to alphanumeric + _ values

    This can be used for measurement names, object names, etc.
    """

    def __init__(self, text, value, *args, **kwargs):
        """Initializer

        text - the explanatory text for the setting UI

        value - the default / initial value

        first_must_be_alpha - True if the first character of the value must
                              be a letter or underbar.
        """
        kwargs = kwargs.copy()
        self.first_must_be_alpha = kwargs.pop("first_must_be_alpha", False)
        super(Alphanumeric, self).__init__(text, value, *args, **kwargs)

    def test_valid(self, pipeline):
        """Restrict names to legal ascii C variables

        First letter = a-zA-Z and underbar, second is that + digit.
        """
        self.validate_alphanumeric_text(self.value, self, self.first_must_be_alpha)

    @staticmethod
    def validate_alphanumeric_text(text, setting, first_must_be_alpha):
        """Validate text as alphanumeric, throwing a validation error if not

        text - text to be validated

        setting - blame this setting on failure

        first_must_be_alpha - True if the first letter has to be alpha or underbar
        """
        if first_must_be_alpha:
            pattern = "^[A-Za-z_][A-Za-z_0-9]*$"
            error = (
                'Names must start with an ASCII letter or underbar ("_")'
                " optionally followed by ASCII letters, underbars or digits."
            )
        else:
            pattern = "^[A-Za-z_0-9]+$"
            error = 'Only ASCII letters, digits and underbars ("_") can be ' "used here"

        match = re.match(pattern, text)
        if match is None:
            raise ValidationError(error, setting)



from ._binary import Binary
from ._binary_matrix import BinaryMatrix
from ._color import Color
from ._coordinates import Coordinates
from ._data_types import DataTypes
from ._delete_setting_event import DeleteSettingEvent
from ._divider import Divider
from ._do_things import DoThings
from ._figure_subscriber import FigureSubscriber
from ._file_collection_display import FileCollectionDisplay
from ._hidden_count import HiddenCount
from ._html_text import HTMLText
from ._image_plane import ImagePlane
from ._joiner import Joiner
from ._measurement import Measurement
from ._path_list_display import PathListDisplay
from ._regexp_text import RegexpText
from ._setting import Setting
from ._settings_group import SettingsGroup
from ._structuring_element import StructuringElement
from ._table import Table
from ._tree_choice import TreeChoice
from ._validation_error import ValidationError



import click

from ..__main__ import pass_environment


@click.group("worker")
@pass_environment
def command(context):
    pass


@command.command("start")
@pass_environment
def start(context):
    pass



from ...utilities.zmq.communicable.reply import Reply


class SharedDictionary(Reply):
    def __init__(self, dictionaries=None):
        Reply.__init__(self, dictionaries=dictionaries)

        if dictionaries is None:
            dictionaries = [{}]



from ._event import Event


class PipelineLoaded(Event):
    """Indicates that the pipeline has been (re)loaded

    """

    def __init__(self):
        super(PipelineLoaded, self).__init__(
            is_pipeline_modification=True, is_image_set_modification=True
        )

    def event_type(self):
        return "PipelineLoaded"



FILTER_CHOICE_NONE = "No filtering"
FILTER_CHOICE_IMAGES = "Images only"
FILTER_CHOICE_CUSTOM = "Custom"
FILTER_CHOICE_ALL = [FILTER_CHOICE_NONE, FILTER_CHOICE_IMAGES, FILTER_CHOICE_CUSTOM]
FILTER_DEFAULT = (
    'and (extension does isimage) (directory doesnot containregexp "[\\\\\\\\/]\\\\.")'
)



from cellprofiler_core.utilities.zmq.communicable._communicable import Communicable
from cellprofiler_core.utilities.zmq.communicable.reply._reply import Reply


class Request(Communicable):
    """A class for making requests and receiving replies across a zmq REQ
    socket.  When communicated through a socket, the class will self-serialize
    any non-callable attributes that do not start with an underscore.

    When received, reply() can be used to send a Reply object to the original
    Request.

    Note that both sides of the connection must have imported the same module
    under the same name defining any objects that will be communicated.

    All subclasses must accept keyword arguments to __init__() corresponding to
    their attributes.
    """

    def __init__(self, **kwargs):
        # all keywords become attributes
        self.__dict__.update(kwargs)
        self._boundary = None

    def send(self, socket):
        Communicable.send(self, socket)
        return Communicable.recv(socket)

    def send_only(self, socket):
        """Send the request but don't perform the .recv

        socket - send on this socket

        First part of a two-part client-side request: send the request
        with an expected .recv, possibly after polling to make the .recv
        non-blocking.
        """
        Communicable.send(self, socket)

    def set_boundary(self, boundary):
        """Set the boundary object to use when sending the reply

        boundary - the reply will be enqueued on this boundary's transmit thread
        """
        self._boundary = boundary

    def reply(self, reply_obj, please_reply=False):
        """send a reply to a request.  If please_reply is True, wait for and
        return a reply to the reply.  Note that that reply should be treated
        like a Request object, i.e., it should be replied to."""
        assert isinstance(
            reply_obj, Reply
        ), "send_reply() called with something other than a Reply object!"
        if self._boundary is None:
            return Communicable.reply(self, reply_obj, please_reply)
        else:
            self._boundary.enqueue_reply(self, reply_obj)



import logging

from .._text import Text
from ..._validation_error import ValidationError


LOGGER = logging.getLogger(__name__)

class Number(Text):
    """A setting that allows only numeric input
    """

    def __init__(self, text, value=0, minval=None, maxval=None, *args, **kwargs):
        if isinstance(value, str):
            text_value = value
            value = self.str_to_value(value)
        else:
            text_value = self.value_to_str(value)

        if kwargs.pop("metadata", False):
            raise ValueError("metadata=True is not a valid argument for a numeric setting.")

        super(Number, self).__init__(text, text_value, *args, **kwargs)
        self.__default = self.str_to_value(text_value)
        self.__minval = minval
        self.__maxval = maxval

    def str_to_value(self, str_value):
        """Return the value of the string passed

        Override this in a derived class to parse the numeric text or
        raise an exception if badly formatted.
        """
        raise NotImplementedError("Please define str_to_value in a subclass")

    def value_to_str(self, value):
        """Return the string representation of the value passed

        Override this in a derived class to convert a numeric value into text
        """
        raise NotImplementedError("Please define value_to_str in a subclass")

    def set_value(self, value):
        """Convert integer to string
        """
        str_value = str(value) if isinstance(value, str) else self.value_to_str(value)
        self.set_value_text(str_value)

    def get_value(self, reraise=False):
        """Return the value of the setting as a float
        """
        return self.__default

    def set_value_text(self, value_text):
        super(Number, self).set_value_text(value_text)
        try:
            self.test_valid(None)
            self.__default = self.str_to_value(value_text)
        except:
            LOGGER.debug("Number set to illegal value: %s" % value_text)

    def set_min_value(self, minval):
        """Programatically set the minimum value allowed"""
        self.__minval = minval

    def set_max_value(self, maxval):
        """Programatically set the maximum value allowed"""
        self.__maxval = maxval

    def get_min_value(self):
        """The minimum value (inclusive) that can legally be entered"""
        return self.__minval

    def get_max_value(self):
        """The maximum value (inclusive) that can legally be entered"""
        return self.__maxval

    min_value = property(get_min_value, set_min_value)
    max_value = property(get_max_value, set_max_value)

    def test_valid(self, pipeline):
        """Return true only if the text value is float
        """
        try:
            value = self.str_to_value(self.value_text)
        except ValueError:
            raise ValidationError("Value not in decimal format", self)
        if self.__minval is not None and self.__minval > value:
            raise ValidationError(
                "Must be at least %s, was %s"
                % (self.value_to_str(self.__minval), self.value_text),
                self,
            )
        if self.__maxval is not None and self.__maxval < value:
            raise ValidationError(
                "Must be at most %s, was %s"
                % (self.value_to_str(self.__maxval), self.value_text),
                self,
            )

    def eq(self, x):
        """Equal if our value equals the operand"""
        return self.value == x



from ._filter_predicate import FilterPredicate
from ._does_predicate import DoesPredicate
from ._does_not_predicate import DoesNotPredicate
from .._file_collection_display import FileCollectionDisplay
from ...constants.image import MD_SIZE_T, MD_SIZE_Z
from ...constants.measurement import C_RGB, C_MONOCHROME


class ImagePredicate(FilterPredicate):
    """A predicate that applies subpredicates to image plane details"""

    IS_COLOR_PREDICATE = FilterPredicate(
        "iscolor",
        "Color",
        lambda plane: (
                plane.color_format == C_RGB
        ),
        [],
        doc="The image is an interleaved color image (for example, a PNG image)",
    )

    IS_MONOCHROME_PREDICATE = FilterPredicate(
        "ismonochrome",
        "Monochrome",
        lambda plane: (
            plane.color_format == C_MONOCHROME
        ),
        [],
        doc="The image is monochrome",
    )

    @staticmethod
    def is_stack(x):
        if MD_SIZE_T in x.metadata and x.metadata[MD_SIZE_T] > 1:
            return True
        if MD_SIZE_Z in x.metadata and x.metadata[MD_SIZE_Z] > 1:
            return True
        return False

    IS_STACK_PREDICATE = FilterPredicate(
        "isstack",
        "Stack",
        lambda x: ImagePredicate.is_stack(x),
        [],
        doc="The image is a Z-stack or movie",
    )

    IS_STACK_FRAME_PREDICATE = FilterPredicate(
        "isstackframe",
        "Stack frame",
        lambda x: x.index is not None,
        [],
        doc="The image is a frame of a movie or a plane of a Z-stack",
    )

    def __init__(self):
        subpredicates = (
            self.IS_COLOR_PREDICATE,
            self.IS_MONOCHROME_PREDICATE,
            self.IS_STACK_PREDICATE,
            self.IS_STACK_FRAME_PREDICATE,
        )
        predicates = [
            pred_class(subpredicates, text)
            for pred_class, text in (
                (DoesPredicate, "Is"),
                (DoesNotPredicate, "Is not"),
            )
        ]

        FilterPredicate.__init__(
            self,
            "image",
            "Image",
            self.fn_filter,
            predicates,
            doc="Filter based on image characteristics",
        )

    @staticmethod
    def fn_filter(node_type__modpath__plane, *args):
        (node_type, modpath, plane) = node_type__modpath__plane
        if node_type == FileCollectionDisplay.NODE_DIRECTORY:
            return None
        if plane is None:
            return None
        return args[0](plane, *args[1:])

    def test_valid(self, pipeline, *args):
        from ...modules.setting_validation import SettingValidation

        self(
            (
                FileCollectionDisplay.NODE_FILE,
                ["/imaging", "test.tif"],
                SettingValidation(),
            ),
            *args,
        )



from ._name import Name
from ...._validation_error import ValidationError


class LabelName(Name):
    """
    A setting that provides an image name
    """

    def __init__(self, text, value="Do not use", *args, **kwargs):
        super(LabelName, self).__init__(text, "objectgroup", value, *args, **kwargs)

    def test_valid(self, pipeline):
        if self.value_text in ["Experiment", "Image", "Relationship"]:
            raise ValidationError(
                "Object names may not be any of %s"
                % (", ".join(["Experiment", "Image", "Relationship"])),
                self,
            )

        super(LabelName, self).test_valid(pipeline)



"""
Readers are modular classes designed to read in image data. Like modules, readers can also be added as plugins
using this template class. Place readers into the plugins directory to load them on startup.

A typical reader plugin will add support for a specific file format or protocol.

N.b. readers should not assume that cellprofiler (and by extension wx/the gui) is always available. In a
headless environment cellprofiler_core may be installed alone.

To add custom file URI schemas to CellProfiler, e.g. "omero:iid=123",
append new entries to the list in cellprofiler_core/constants/image.py.
This will allow the main file list to accept those URIs.

If you need to add menu entries to the GUI, access cellprofiler.gui.plugins_menu and add
entries to the container within. Further instructions are within that file.
"""
import uuid

from abc import ABC, abstractmethod


class Reader(ABC):
    """
    Derive from this abstract Reader class to create your own image reader in Python

    You need to implement the methods below in the derived class.

    Use this block of text to describe your reader to the user.
    """

    def __init__(self, image_file):
        """
        Your init should define a variable_revision_number representing the reader version.
        Defining reader_name also allows you to register the reader with a custom name,
        otherwise it'll match the class name.
        """
        self.file = image_file
        self.id = uuid.uuid4()

    @property
    @abstractmethod
    def supported_filetypes(self):
        # This should be a class property. Give the reader a set of supported filetypes (extensions).
        pass

    @property
    @abstractmethod
    def supported_schemes(self):
        # This should be a class property. Give the reader a set of supported schemes (e.g. file, https, s3, ...).
        pass

    @property
    @abstractmethod
    def variable_revision_number(self):
        # This should be a class property. Give the reader a version number (int).
        pass

    @property
    @abstractmethod
    def reader_name(self):
        # This should be a class property. Give the reader a human-readable name (string).
        pass

    @abstractmethod
    def read(self,
             wants_metadata_rescale=False,
             series=None,
             index=None,
             c=None,
             z=None,
             t=None,
             xywh=None,
             channel_names=None,
             ):
        """Read a single plane from the image file. Mimics the Bioformats API
        :param wants_metadata_rescale: if `True`, return a tuple of image and a
               tuple of (min, max) for range values of image dtype gathered from
               file metadata; if `False`, returns only the image
        :param c: read from this channel. `None` = read color image if multichannel
            or interleaved RGB.
        :param z: z-stack index
        :param t: time index
        :param series: series for ``.flex`` and similar multi-stack formats
        :param index: if `None`, fall back to ``zct``, otherwise load the indexed frame
        :param xywh: a (x, y, w, h) tuple
        :param channel_names: provide the channel names for the OME metadata

        Should return a data array with channel order X, Y, (C)
        """
        return None

    def read_volume(self,
                    wants_metadata_rescale=False,
                    series=None,
                    c=None,
                    z=None,
                    t=None,
                    xywh=None,
                    channel_names=None,
                    ):
        """Read a series of planes from the image file. Mimics the Bioformats API
        :param wants_metadata_rescale: if `True`, return a tuple of image and a
               tuple of (min, max) for range values of image dtype gathered from
               file metadata; if `False`, returns only the image
        :param c: read from this channel. `None` = read color image if multichannel
            or interleaved RGB.
        :param z: z-stack index
        :param t: time index
        n.b. either z or t should be "None" to specify which channel to read across.
        :param series: series for ``.flex`` and similar multi-stack formats
        :param xywh: a (x, y, w, h) tuple
        :param channel_names: provide the channel names for the OME metadata

        Should return a data array with channel order Z, X, Y, (C)
        """
        raise NotImplementedError(f"This reader ({self.reader_name}) does not support 3D reading.")

    @classmethod
    @abstractmethod
    def supports_format(cls, image_file, allow_open=False, volume=False):
        """This function needs to evaluate whether a given ImageFile object
        can be read by this reader class.

        Return value should be an integer representing suitability:
        -1 - 'I can't read this at all'
        1 - 'I am the one true reader for this format, don't even bother checking any others'
        2 - 'I am well-suited to this format'
        3 - 'I can read this format, but I might not be the best',
        4 - 'I can give it a go, if you must'

        The allow_open parameter dictates whether the reader is permitted to read the file when
        making this decision. If False the decision should be made using file extension only.
        Any opened files should be closed before returning. For now readers are selected with
        allow_open disabled, but this may change in the future.

        The volume parameter specifies whether the reader will need to return a 3D array.
        ."""
        return -1

    @classmethod
    def clear_cached_readers(cls):
        # This should clear any cached reader objects if your class stores unused readers.
        pass

    @classmethod
    def supports_url(cls):
        # This function defines whether the reader class supports reading data directly from a URL.
        # If False, CellProfiler will download compatible web-based images to a temporary directory.
        # If True, the reader will be passed the source URL.
        return False

    @abstractmethod
    def close(self):
        # If your reader opens a file, this needs to release any active lock,
        pass

    @abstractmethod
    def get_series_metadata(self):
        """Should return a dictionary with the following keys:
        Key names are in cellprofiler_core.constants.image
        MD_SIZE_S - int reflecting the number of series
        MD_SIZE_X - list of X dimension sizes, one element per series.
        MD_SIZE_Y - list of Y dimension sizes, one element per series.
        MD_SIZE_Z - list of Z dimension sizes, one element per series.
        MD_SIZE_C - list of C dimension sizes, one element per series.
        MD_SIZE_T - list of T dimension sizes, one element per series.
        MD_SERIES_NAME - [Optional] list of human-readable series names, one string per series.
                                    Must not contain '|' as this is used as a separator for storage.
        """
        pass

    @staticmethod
    def get_settings():
        """
        This function should return a list of settings objects configurable
        by the reader class. Each entry is a tuple. This list should not
        include the default '.enabled' key, used to disable readers.

        Tuple Format: (config key, name, description, type, default)

        config key - A name for the key in internal storage.
        Slashes should not be used in key names. The reader's name will
        be prefixed automatically.

        name - a human-readable name to be shown to the user

        description - extra text to describe the setting

        type - one of str, bool, float or int. These types are supported by
        wx config files.

        default - value to use if no existing config exists

        :return: list of setting tuples
        """
        return []



from ...utilities.zmq.communicable.request import AnalysisRequest


class MeasurementsReport(AnalysisRequest):
    def __init__(self, analysis_id, buf, image_set_numbers=None):
        AnalysisRequest.__init__(
            self, analysis_id, buf=buf, image_set_numbers=image_set_numbers
        )
        if image_set_numbers is None:
            image_set_numbers = []



"""analysis_worker.py - Run pipelines on imagesets to produce measurements.

The analysis worker listens on a ZMQ port for work announcements. It then
requests jobs from the announcer and executes them. As an application,
the analysis worker runs three threads:

* Main thread - spawns the worker and monitor threads and enters a run loop.
                The run loop is needed on OS/X in order to process the UI.
                The UI is needed by ImageJ 1.0 which starts AWT. The main thread
                issues a stop notification to the worker thread after exiting
                the run loop.

* Worker thread - listens for jobs and processes them until it receives a stop
                  notification from the main thread.

* Monitor thread - reads from STDIN. If the parent process closes STDIN,
                   the read call throws an exception and the monitor thread
                   stops the main thread's run loop.
"""

import logging
import os
import sys
import psutil

import importlib.resources

from cellprofiler_core.constants.worker import (
    DEADMAN_START_ADDR,
    DEADMAN_START_MSG,
    NOTIFY_STOP,
    all_measurements, NOTIFY_ADDR,
)
from cellprofiler_core.preferences import set_always_continue, set_conserve_memory
from cellprofiler_core.worker._worker import Worker
from cellprofiler_core.utilities.logging import set_log_level


LOGGER = logging.getLogger(__name__)

"""Set the log level through the environment by specifying AW_LOG_LEVEL"""
AW_LOG_LEVEL = "AW_LOG_LEVEL"

knime_bridge_address = None
notify_address = None
analysis_id = None
work_server_address = None


def aw_parse_args():
    """Parse the application arguments into setup parameters"""
    from cellprofiler_core.preferences import (
        set_headless,
        set_awt_headless,
        set_plugin_directory,
    )
    import argparse

    global analysis_id
    global work_server_address
    global notify_address
    global knime_bridge_address

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--analysis-id",
        dest="analysis_id",
        help="Unique ID for the analysis to be performed",
        default=None,
    )
    parser.add_argument(
        "--work-server",
        dest="work_server_address",
        help="ZMQ port where work requests should be sent",
        default=None,
    )
    parser.add_argument(
        "--notify-server",
        dest="notify_address",
        help="ZMQ port where continue/shutdown notifications are published",
        default=None,
    )
    parser.add_argument(
        "--log-level",
        dest="log_level",
        default=os.environ.get(AW_LOG_LEVEL, logging.INFO),
        help=(
            "Set the verbosity for logging messages: "
            + ("%d or %s for debugging, " % (logging.DEBUG, "DEBUG"))
            + ("%d or %s for informational, " % (logging.INFO, "INFO"))
            + ("%d or %s for warning, " % (logging.WARNING, "WARNING"))
            + ("%d or %s for error, " % (logging.ERROR, "ERROR"))
            + ("%d or %s for critical, " % (logging.CRITICAL, "CRITICAL"))
            + ("%d or %s for fatal." % (logging.FATAL, "FATAL"))
            + " Otherwise, the argument is interpreted as the file name of a log configuration file (see http://docs.python.org/library/logging.config.html for file format)"
        ),
    )
    parser.add_argument(
        "--plugins-directory",
        dest="plugins_directory",
        help="Folder containing the CellProfiler plugin modules needed by client pipelines",
        default=None,
    )
    parser.add_argument(
        "--conserve-memory",
        dest="conserve_memory",
        default=None,
        help="CellProfiler will attempt to release unused memory after each image set.",
    )
    parser.add_argument(
        "--jvm-heap-size",
        dest="jvm_heap_size",
        default=None,
        help=(
            "This is the amount of memory reserved for the "
            "Java Virtual Machine (similar to the java -Xmx switch)."
            "Example formats: 512000k, 512m, 1g"
        ),
    )
    parser.add_argument(
        "--knime-bridge-address",
        dest="knime_bridge_address",
        help="Open up a port to handle the Knime bridge protocol",
        default=None,
    )
    parser.add_argument(
        "--always-continue",
        dest="always_continue",
        help="Don't stop the analysis when an image set raises an error",
        default=None
    )

    args = parser.parse_args()
    
    set_log_level(args.log_level, subprocess=True)

    set_awt_headless(True)
    set_headless()

    # NOTE: don't call fill_readers here,
    # woker needs to do PipelinePreferences req then set_preferences_from_dict,
    # which it does later in worker.do_job

    if not args.work_server_address and args.work_server_address and \
            args.analysis_id:
        parser.print_help()
        sys.exit(1)
    analysis_id = args.analysis_id
    notify_address = args.notify_address
    work_server_address = args.work_server_address
    knime_bridge_address = args.knime_bridge_address

    #
    # Set up the headless plugins directories before doing
    # anything so loading will get them
    #
    if args.plugins_directory is not None:
        set_plugin_directory(args.plugins_directory, globally=False)
    if args.conserve_memory is not None:
        set_conserve_memory(args.conserve_memory, globally=False)
    if args.always_continue is not None:
        set_always_continue(args.always_continue, globally=False)
    else:
        LOGGER.warning("Plugins directory not set")


if __name__ == "__main__":
    #
    # Sorry to put ugliness so early: The process inherits file descriptors
    # from the parent. Windows doesn't let you selectively inherit file
    # descriptors, so we close them here.
    #
    try:
        maxfd = os.sysconf("SC_OPEN_MAX")
    except:
        maxfd = 256

    proc = psutil.Process()

    # This is a hacky solution to an annoying problem with vscode debugging:
    # we want to set breakpoints in the debugger,
    # but one of the file descriptors inhertited by the child
    # corresponds to a TCP/IPv4 connection attached to the debugger,
    # so closing it would cause the debugger to detach from the subprocess.
    # AFAICT the debugger fd is always 3 or 4, and is not associated with
    # any device + inode combo (both 0, ie null), so
    # if we find that, skip closing fd 4, else close all
    try:
        stat3 = os.fstat(3)
        stat4 = os.fstat(4)
        if stat3.st_ino == 0 and stat3.st_dev == 0:
            os.closerange(4, maxfd)
        elif stat4.st_ino == 0 and stat4.st_dev == 0:
            os.close(3)
            os.closerange(5, maxfd)
        else:
            os.closerange(3, maxfd)
    except OSError:
        os.closerange(3, maxfd)

    if not hasattr(sys, "frozen"):
        # In the development version, maybe the bioformats package is installed?
        # Add the root to the pythonpath
        root = os.path.split(os.path.split(__file__)[0])[0]
        sys.path.insert(0, root)

    aw_parse_args()

import time
import threading
import zmq

#
# CellProfiler expects NaN as a result during calculation
#
import numpy as np

np.seterr(all="ignore")


# to guarantee closing of measurements, we store all of them in a WeakSet, and
# close them on exit.


def main():
    #
    # For OS/X set up the UI elements that users expect from
    # an app.
    #
    if sys.platform == "darwin":
        import os.path

        icon_path = str(importlib.resources.files("cellprofiler").joinpath(
            "data", "icons", "CellProfiler.png"
        ))
        os.environ["APP_NAME_%d" % os.getpid()] = "CellProfilerWorker"
        os.environ["APP_ICON_%d" % os.getpid()] = icon_path

    # Start the JVM
    from cellprofiler_core.utilities.java import start_java, stop_java
    with zmq.Context() as the_zmq_context:
        the_zmq_context.setsockopt(zmq.LINGER, 0)
        deadman_start_socket = the_zmq_context.socket(zmq.PAIR)
        deadman_start_socket.bind(DEADMAN_START_ADDR)

        # Start the deadman switch thread.
        monitor_thread = threading.Thread(
            target=monitor_keepalive,
            args=(the_zmq_context, notify_address, ),
            name="heartbeat_monitor thread",
            daemon=True
        )
        monitor_thread.start()

        deadman_start_socket.recv()
        deadman_start_socket.close()

        from cellprofiler.knime_bridge import KnimeBridgeServer

        with Worker(the_zmq_context, analysis_id, work_server_address, notify_address) as worker:
            worker_thread = threading.Thread(
                target=worker.run,
                name="WorkerThread",
                daemon=True,
            )
            worker_thread.start()
            with KnimeBridgeServer(
                the_zmq_context, knime_bridge_address, NOTIFY_ADDR, NOTIFY_STOP
            ):
                worker_thread.join()
            the_zmq_context.destroy(linger=0)
        LOGGER.debug("Worker thread joined")
        #
        # Shutdown - need to handle some global cleanup here
        #
        try:
            stop_java()
        except:
            LOGGER.warning("Failed to stop the JVM", exc_info=True)


def monitor_keepalive(context, keepalive_address):
    """The keepalive socket should send a regular heartbeat telling workers
    to stay alive. This will stop if the parent process crashes.
    If we see no keepalive message in 15 seconds we take that as a bad
    omen and attempt to gracefully shut down the worker processes.

    The same socket also broadcasts the shutdown notification, so we watch
    for that too. The main worker thread should pick that up and shut down
    the next time it makes a request, but this can take a while. Therefore we
    also apply a timeout and cease work manually if the processes don't
    respond to the kill signal quickly enough.

    Note that if CellProfiler is paused in debug mode during an analysis, this
    can cause the workers to exit. Use test mode for testing or only debug on
    the worker threads."""
    keepalive_socket = context.socket(zmq.SUB)
    keepalive_socket.connect(keepalive_address)
    keepalive_socket.setsockopt(zmq.SUBSCRIBE, b"")

    # Send a message back when we've set this thread up
    deadman_socket = context.socket(zmq.PAIR)
    deadman_socket.connect(DEADMAN_START_ADDR)
    deadman_socket.send(DEADMAN_START_MSG)
    deadman_socket.close()

    missed = 0
    while missed < 3:
        event = keepalive_socket.poll(5000)
        if not event:
            missed += 1
            LOGGER.warning(f"Worker failed to receive communication for"
                            f" {5 * missed} seconds")
        else:
            missed = 0
            msg = keepalive_socket.recv()
            if msg == NOTIFY_STOP:
                break
    # Stop has been called, we must manually close our sockets before the
    # main thread can exit.
    keepalive_socket.close()
    if missed >= 3:
        # Parent is dead, hard kill
        LOGGER.critical("Worker stopped receiving communication from "
                         "CellProfiler, shutting down now")
    else:
        # Stop signal captured, give some time to shut down gracefully.
        time.sleep(10)
    LOGGER.info("Cancelling worker")
    # hard exit after 10 seconds unless app exits

    for m in all_measurements:
        try:
            m.close()
        except:
            pass
    LOGGER.error("Worker failed to stop gracefully, forcing exit now")
    os._exit(0)


if __name__ == "__main__":
    main()
    sys.exit(0)



import csv
import logging
import numbers
import os
import re
import string
import time
import urllib.request

from ..constants.image import MD_SIZE_C, MD_SIZE_T, MD_SIZE_Z, MD_SIZE_X, MD_SIZE_Y
from ..constants.measurement import COLTYPE_FLOAT, C_Z, C_T, C_C, FTR_WELL, ROW_KEYS, COL_KEYS
from ..constants.measurement import COLTYPE_INTEGER
from ..constants.measurement import COLTYPE_VARCHAR
from ..constants.measurement import COLTYPE_VARCHAR_FILE_NAME
from ..constants.measurement import C_FRAME
from ..constants.measurement import C_METADATA
from ..constants.measurement import C_SERIES
from ..constants.measurement import RESERVED_METADATA_KEYS
from ..constants.module import FILTER_RULES_BUTTONS_HELP
from ..constants.module import PROTIP_RECOMMEND_ICON
from ..constants.modules.metadata import COL_INDEX
from ..constants.modules.metadata import DEFAULT_METADATA_TAGS
from ..constants.modules.metadata import CSV_JOIN_NAME
from ..constants.modules.metadata import IPD_JOIN_NAME
from ..constants.modules.metadata import COL_PATH
from ..constants.modules.metadata import COL_URL
from ..constants.modules.metadata import COL_SERIES
from ..constants.modules.metadata import DTC_ALL
from ..constants.modules.metadata import DTC_CHOOSE
from ..constants.modules.metadata import DTC_TEXT
from ..constants.modules.metadata import F_ALL_IMAGES
from ..constants.modules.metadata import F_FILTERED_IMAGES
from ..constants.modules.metadata import IDX_EXTRACTION_METHOD_COUNT
from ..constants.modules.metadata import IDX_EXTRACTION_METHOD_COUNT_V1
from ..constants.modules.metadata import IDX_EXTRACTION_METHOD_COUNT_V2
from ..constants.modules.metadata import IDX_EXTRACTION_METHOD_COUNT_V3
from ..constants.modules.metadata import IDX_EXTRACTION_METHOD_V1
from ..constants.modules.metadata import IDX_EXTRACTION_METHOD_V2
from ..constants.modules.metadata import LEN_EXTRACTION_METHOD
from ..constants.modules.metadata import LEN_EXTRACTION_METHOD_V1
from ..constants.modules.metadata import XM_FILE_NAME
from ..constants.modules.metadata import XM_FOLDER_NAME
from ..constants.modules.metadata import X_ALL_EXTRACTION_METHODS
from ..constants.modules.metadata import X_AUTOMATIC_EXTRACTION
from ..constants.modules.metadata import X_IMPORTED_EXTRACTION
from ..constants.modules.metadata import X_MANUAL_EXTRACTION
from ..module import Module
from ..pipeline import Pipeline
from ..preferences import ABSOLUTE_FOLDER_NAME
from ..preferences import URL_FOLDER_NAME
from ..preferences import report_progress
from ..setting import Binary
from ..setting import DataTypes
from ..setting import Divider
from ..setting import FileCollectionDisplay
from ..setting import HiddenCount
from ..setting import Joiner
from ..setting import RegexpText
from ..setting import SettingsGroup
from ..setting import Table
from ..setting import ValidationError
from ..setting.choice import Choice
from ..setting.do_something import DoSomething
from ..setting.do_something import RemoveSettingButton
from ..setting.filter import (
    Filter,
    FilePredicate,
    DirectoryPredicate,
    ExtensionPredicate,
)
from cellprofiler_core.setting.text import Directory
from cellprofiler_core.setting.text import Filename
from cellprofiler_core.utilities.image import generate_presigned_url
from cellprofiler_core.utilities.image import image_resource
from cellprofiler_core.utilities.measurement import find_metadata_tokens


LOGGER = logging.getLogger(__name__)

__doc__ = """\
Metadata
========

The **Metadata** module allows you to extract and associate metadata
with your images. The metadata can be extracted from the image file
itself, from a part of the file name or location, and/or from a text
file you provide.

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          NO
============ ============ ===============

What is “metadata”?
^^^^^^^^^^^^^^^^^^^

The term *metadata* refers to “data about data.” For many assays,
metadata is important in the context of tagging images with various
attributes, which can include (but is not limited to) items such as the
following:

-  The row and column of the microtiter plate that the image was
   acquired from.
-  The experimental treatment applied to the well that the image was
   acquired from.
-  The number of timepoints or channels contained in the image file.
-  The image type, i.e., RGB, indexed or separate channels.
-  The height and width of an image, in pixels.

It can be helpful to inform CellProfiler about certain metadata in order
to define a specific relationship between the images and the associated
metadata. For instance:

-  You want images with a common tag to be matched together so they are
   processed together during the pipeline run. E.g., the filenames for
   fluorescent DAPI and GFP images contain different tags indicating the
   wavelength but share ‘\_s1’ in the filename if they were acquired
   from site #1, ‘\_s2’ from site #2, and so on.
-  You want certain information attached to the output measurements and
   filenames for annotation or sample-tracking purposes. E.g., some
   images are to be identified as acquired from DMSO treated wells,
   whereas others were collected from wells treated with Compound 1, 2,…
   and so forth.

The underlying assumption in matching metadata values to image sets is
that there is an exact pairing (i.e., a one-to-one match) for a given
combination of metadata tags. A common example is that for a two-channel
microtiter plate assay, the values of the plate, well, and site tags
from one channel get matched uniquely to the plate, well, and site tag
values from the other channel.

What are the inputs?
^^^^^^^^^^^^^^^^^^^^

If you do not have metadata that is relevant to your analysis, you can
leave this module in the default "*No*" setting, and continue on to the
**NamesAndTypes** module If you do have relevant metadata, the
**Metadata** module receives the file list produced by the **Images**
module. It then associates information to each file in the file list,
which can be obtained from several sources:

-  From the image file name or location (e.g., as assigned by a
   microscope). In this case, you will provide the text search pattern
   to obtain this information.
-  In a text file created and filled out by you or a laboratory
   information management system. In this case, you will point the
   module to the location of this file.

You can extract metadata from all images loaded via the **Images**
module, or a subset of them by using rules to filter the list.

What do I get as output?
^^^^^^^^^^^^^^^^^^^^^^^^

The final product of the **Metadata** module is a list of files from the
**Images** module, accompanied by the associated metadata retrieved
from the source(s) provided and matched to the desired images.

As you are extracting metadata from your various sources, you can click
the “Update” button below the divider to display a table of results
using the current settings. Each row corresponds to an image file from
the **Images** module, and the columns display the metadata obtained for
each tag specified. You can press this button as many times as needed to
display the most current metadata obtained.

.. image:: {METADATA_DISPLAY_TABLE}
   :width: 100%

Some downstream use cases for metadata include the following:

-  If the metadata establishes how channels are related to one another,
   you can use them in the **NamesAndTypes** module to aid in creating
   an image set.
-  If the images need to be further sub-divided into groups of images
   that share a common metadata value, the **Groups** module can be used
   to specify which metadata is needed for this purpose.
-  You can also use the numerical values of pieces of metadata in later modules.
   Since the metadata is stored as an image measurement and can be
   assigned as an integer or floating-point number, any module which
   allows measurements as input can make use of it.
-  Several modules are also capable of using metadata for more specific
   purposes. Refer to the module setting help for additional information
   on how to use them in the context of the specific module.

If the metadata originates from an external source such as a CSV, there
are some caveats in the cases when metadata is either missing or
duplicated for the referenced images; see the **NamesAndTypes** module
for more details.

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *Metadata:* The prefix of each metadata tag in the per-image table.
""".format(
    **{"METADATA_DISPLAY_TABLE": image_resource("Metadata_ExampleDisplayTable.png")}
)


class Metadata(Module):
    variable_revision_number = 7
    module_name = "Metadata"
    category = "File Processing"

    CSV_JOIN_NAME = "CSV Metadata"
    IPD_JOIN_NAME = "Image Metadata"

    def create_settings(self):
        self.pipeline = None
        self.filtered_file_list = None
        # Records whether upgrade process removed an automatic extraction group
        self.removed_automatic_extraction = False

        module_explanation = [
            "The %s module optionally allows you to extract information"
            % self.module_name,
            "describing your images (i.e, metadata) which will be stored along",
            "with your measurements. This information can be contained in the",
            "file name and/or location, or in an external file.",
        ]
        self.set_notes([" ".join(module_explanation)])

        self.wants_metadata = Binary(
            "Extract metadata?",
            False,
            doc="""\
Select "*{YES}*" if your file or path names contain
information (i.e., metadata) you would like to extract and store along
with your measurements. See the main module help for more details.
""".format(
                **{"YES": "Yes"}
            ),
        )

        self.extraction_methods = []

        self.add_extraction_method(False)

        self.extraction_method_count = HiddenCount(
            self.extraction_methods, "Extraction method count"
        )

        self.add_extraction_method_button = DoSomething(
            "", "Add another extraction method", self.add_extraction_method
        )

        self.dtc_divider = Divider()

        self.data_type_choice = Choice(
            "Metadata data type",
            DTC_ALL,
            tooltips={
                DTC_TEXT: "Save all metadata as text",
                DTC_CHOOSE: "Choose the data type (text or numeric) for each metadata category",
            },
            doc="""\
Metadata can be stored as either a text or numeric value:

-  *{DTC_TEXT}:* Save all metadata item as text.
-  *{DTC_CHOOSE}:* Choose the data type separately for each metadata
   entry. An example of when this approach would be necessary would be
   if a whole filename is captured as metadata but the file name is
   numeric, e.g., “0001101”. In this situation, if the file name needs
   to be used for an arithmetic calculation or index, the name would
   need to be converted to a number and you would select “Integer” as
   the data type. On the other hand, if it important that the leading
   zeroes be retained, setting it to an integer would remove them upon
   conversion to a number. In this case, storing the metadata values as
   “Text” would be more appropriate.
""".format(
                **{"DTC_CHOOSE": DTC_CHOOSE, "DTC_TEXT": DTC_TEXT}
            ),
        )

        self.data_types = DataTypes(
            "Metadata types",
            name_fn=self.get_metadata_keys,
            doc="""\
*(Used only when “{DTC_CHOOSE}” is selected for the metadata data type)*

This setting determines the data type of each metadata field when
stored as a measurement.

-  *Text:* Save the metadata as text.
-  *Integer:* Save the metadata as an integer.
-  *Float:* Save the metadata as a decimal number.
-  *None:* Do not save the metadata as a measurement.

If you are using the metadata to match images to create your image set,
the choice of metadata type here will determine the order of matching.
See **NamesAndTypes** for more details.
""".format(
                **{"DTC_CHOOSE": DTC_CHOOSE}
            ),
        )

        self.table = Table(
            "",
            use_sash=True,
            corner_button=dict(
                fn_clicked=self.update_table,
                label="Update",
                tooltip="Update the metadata table",
            ),
        )

    def add_extraction_method(self, can_remove=True):
        group = SettingsGroup()

        if can_remove:
            group.append("divider", Divider())

        group.append(
            "extraction_method",
            Choice(
                "Metadata extraction method",
                X_ALL_EXTRACTION_METHODS,
                X_MANUAL_EXTRACTION,
                doc="""\
Metadata can be stored in either or both of two ways:

-  *Internally:* This method is often through the file naming, directory
   structuring, or the file header information.
-  *Externally:* This is through an external index, such as spreadsheet
   or database of some kind.

The **Metadata** module can extract internal or external metadata from
the images in any of three ways:

-  *{X_MANUAL_EXTRACTION}*: This approach retrieves information based
   on the file nomenclature and/or location. A special syntax called
   “regular expressions” is used to match text patterns in the file name
   or path, and then assign this text as metadata for the images you
   specify. The tag for each metadata is assigned a name that is
   meaningful to you.

   |image0|  *When would you want to use this
   option?* If you want to take advantage of the fact that acquisition
   software often automatically assigns a regular nomenclature to the
   filenames or the containing folders. Alternately, the researcher
   acquiring the images may also have a specific nomenclature they
   adhere to for bookkeeping purposes.
-  *{X_IMPORTED_EXTRACTION}*: This option retrieves metadata from a
   comma-delimited file (known as a CSV file, for comma-separated
   values) of information; you will be prompted to specify the location
   of the CSV file. You can create such a file using a spreadsheet
   program such as Microsoft Excel.

   Please make sure your metadata file name does not contain parentheses
   or spaces.
   In general, avoid having spaces in your metadata file as it could cause 
   errors in later analysis steps. You can remove spaces using a spreadsheet
   program such as Microsoft Excel and using Find and Replace function.

   |image1|  *When would you want to
   use this option?* You have information curated in software that
   allows for export to a spreadsheet. This is commonly the case for
   laboratories that use data management systems that track samples and
   acquisition.

Specifics on the metadata extraction options are described below. Any or
all of these options may be used at time; press the “Add another
extraction method” button to add more.

.. |image0| image:: {PROTIP_RECOMMEND_ICON}
.. |image1| image:: {PROTIP_RECOMMEND_ICON}
""".format(
                    **{
                        "X_IMPORTED_EXTRACTION": X_IMPORTED_EXTRACTION,
                        "X_MANUAL_EXTRACTION": X_MANUAL_EXTRACTION,
                        "PROTIP_RECOMMEND_ICON": PROTIP_RECOMMEND_ICON,
                    }
                ),
            ),
        )

        group.append(
            "source",
            Choice(
                "Metadata source",
                [XM_FILE_NAME, XM_FOLDER_NAME],
                doc="You can extract the metadata from the image's file name or from its folder name.",
            ),
        )

        group.append(
            "file_regexp",
            RegexpText(
                "Regular expression to extract from file name",
                "^(?P<Plate>.*)_(?P<Well>[A-P][0-9]{2})_s(?P<Site>[0-9])_w(?P<ChannelNumber>[0-9])",
                get_example_fn=self.example_file_fn,
                doc="""\
*(Used only if you want to extract metadata from the file name)*

The regular expression to extract the metadata from the file name is
entered here. Please see the general module help for more information on
construction of a regular expression.

Clicking the magnifying glass icon to the right will bring up a tool for
checking the accuracy of your regular expression. The regular expression
syntax can be used to name different parts of your expression. The
syntax *(?P<fieldname>expr)* will extract whatever matches *expr* and
assign it to the measurement *fieldname* for the image.

For instance, a researcher uses plate names composed of a string of
letters and numbers, followed by an underscore, then the well,
followed by another underscore, followed by an “s” and a digit
representing the site taken within the well (e.g.,
*TE12345\_A05\_s1.tif*). The following regular expression will capture
the plate, well, and site in the fields “Plate”, “Well”, and “Site”:

+----------------------------------------------------------------+------------------------------------------------------------------+
| ^(?P<Plate>.\*)\_(?P<Well>[A-P][0-9]{1,2})\_s(?P<Site>[0-9])                                                                      |
+----------------------------------------------------------------+------------------------------------------------------------------+
| ^                                                              | Start only at beginning of the file name                         |
+----------------------------------------------------------------+------------------------------------------------------------------+
| (?P<Plate>                                                     | Name the captured field *Plate*                                  |
+----------------------------------------------------------------+------------------------------------------------------------------+
| .\*)                                                           | Capture as many characters as follow                             |
+----------------------------------------------------------------+------------------------------------------------------------------+
| \_                                                             | Discard the underbar separating plate from well                  |
+----------------------------------------------------------------+------------------------------------------------------------------+
| (?P<Well>                                                      | Name the captured field *Well*                                   |
+----------------------------------------------------------------+------------------------------------------------------------------+
| [A-P]                                                          | Capture exactly one letter between A and P                       |
+----------------------------------------------------------------+------------------------------------------------------------------+
| [0-9]{1,2} )                                                   | Capture one or two digits that follow                            |
+----------------------------------------------------------------+------------------------------------------------------------------+
| \_s                                                            | Discard the underbar followed by *s* separating well from site   |
+----------------------------------------------------------------+------------------------------------------------------------------+
| (?P<Site>                                                      | Name the captured field *Site*                                   |
+----------------------------------------------------------------+------------------------------------------------------------------+
| [0-9])                                                         | Capture one digit following                                      |
+----------------------------------------------------------------+------------------------------------------------------------------+

The regular expression can be typed in the upper text box, with a sample
file name given in the lower text box. Provided the syntax is correct,
the corresponding fields will be highlighted in the same color in the
two boxes. Press *Submit* to enter the typed regular expression.

Note that if you use the special fieldnames *<WellColumn>* and
*<WellRow>* together, LoadImages will automatically create a *<Well>*
metadata field by joining the two fieldname values together. For
example, if *<WellRow>* is “A” and *<WellColumn>* is “01”, a field
*<Well>* will be “A01”. This is useful if your well row and column names
are separated from each other in the filename, but you want to retain
the standard well nomenclature.
""",
            ),
        )

        group.append(
            "folder_regexp",
            RegexpText(
                "Regular expression to extract from folder name",
                "(?P<Date>[0-9]{4}_[0-9]{2}_[0-9]{2})$",
                get_example_fn=self.example_directory_fn,
                guess=RegexpText.GUESS_FOLDER,
                doc="""\
*(Used only if you want to extract metadata from the path)*

Enter the regular expression for extracting the metadata from the
path. Note that this field is available whether you have selected
*Text-Regular expressions* to load the files or not.

Clicking the magnifying glass icon to the right will bring up a tool
that will allow you to check the accuracy of your regular expression.
The regular expression syntax can be used to name different parts of
your expression. The syntax *(?P<fieldname>expr)* will extract whatever
matches *expr* and assign it to the image’s *fieldname* measurement.

For instance, a researcher uses folder names with the date and
subfolders containing the images with the run ID (e.g.,
*./2009\_10\_02/1234/*) The following regular expression will capture
the plate, well, and site in the fields *Date* and *Run*:

+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| .\*[\\\\\\\\/](?P<Date>.\*)[\\\\\\\\/](?P<Run>.\*)$                                                                                                                                                                                          |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| .\*[\\\\\\\\/]                                          | Skip characters at the beginning of the pathname until either a slash (/) or backslash (\\\\) is encountered (depending on the operating system)                                 |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| (?P<Date>                                           | Name the captured field *Date*                                                                                                                                                 |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| .\*)                                                | Capture as many characters that follow                                                                                                                                         |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| [\\\\\\\\/]                                             | Discard the slash/backslash character                                                                                                                                          |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| (?P<Run>                                            | Name the captured field *Run*                                                                                                                                                  |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| .\* )                                               | Capture as many characters as follow                                                                                                                                           |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
| $                                                   | The *Run* field must be at the end of the path string, i.e., the last folder on the path. This also means that the Date field contains the parent folder of the Date folder.   |
+-----------------------------------------------------+--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
""",
            ),
        )

        group.append(
            "filter_choice",
            Choice(
                "Extract metadata from",
                [F_ALL_IMAGES, F_FILTERED_IMAGES],
                doc="""\
Choose whether you want to extract metadata from all of the images
chosen by the **Images** module or a subset of the images.

This setting controls how different image types (e.g., an image of the
GFP stain and a brightfield image) have different metadata extracted.
There are two choices:

-  *{F_ALL_IMAGES}*: Extract metadata from all images specified in
   **Images**. This is the simplest choice and the appropriate one if
   you have only one kind of image (or only one image). CellProfiler
   will extract metadata from all images using the same method per
   iteration.
-  *{F_FILTERED_IMAGES}*: Extract metadata depending on specific
   file attributes. This is the appropriate choice if more than one
   image was taken of each imaging site. You can specify distinctive
   criteria for each image subset with matching metadata.
""".format(
                    **{
                        "F_ALL_IMAGES": F_ALL_IMAGES,
                        "F_FILTERED_IMAGES": F_FILTERED_IMAGES,
                    }
                ),
            ),
        )

        group.append(
            "filter",
            Filter(
                "Select the filtering criteria",
                [FilePredicate(), DirectoryPredicate(), ExtensionPredicate(), ],
                'and (file does contain "")',
                doc="""\
Select "*{YES}*" to display and use rules to select files for metadata
extraction.

{FILTER_RULES_BUTTONS_HELP}
""".format(
                    **{
                        "FILTER_RULES_BUTTONS_HELP": FILTER_RULES_BUTTONS_HELP,
                        "YES": "Yes",
                    }
                ),
            ),
        )

        group.append(
            "csv_location",
            Directory(
                "Metadata file location",
                support_urls=True,
                doc="""\
*(Used only if you want to extract metadata from a file)*

The file containing the metadata must be a comma-delimited file (CSV).
You can create or edit such a file using a spreadsheet program such as
Microsoft Excel.

The CSV file needs to conform to the following format:

-  Each column describes one type of metadata.
-  Each row describes the metadata for one image site.
-  The column headers are uniquely named. You can optionally prepend
   “Metadata\_” to the header name in order to insure that it is
   interpreted correctly.
-  The CSV must be plain text, i.e., without hidden file encoding
   information. If using Excel on a Mac to edit the file, choose to save
   the file as “Windows CSV” or “Windows Comma Separated”.

The file must be saved as plain text, i.e., without hidden file encoding
information. If using Excel on a Mac to edit the file, choose to save
the file as “Windows CSV” or “Windows Comma Separated”.
""",
            ),
        )

        group.append(
            "csv_filename",
            Filename(
                "Metadata file name",
                "None",
                browse_msg="Choose CSV file",
                exts=[("Data file (*.csv)", "*.csv")],
                doc="Provide the file name of the CSV file containing the metadata you want to load.",
                get_directory_fn=group.csv_location.get_absolute_path,
                set_directory_fn=lambda path: group.csv_location.join_parts(
                    *group.csv_location.get_parts_from_path(path)
                ),
            ),
        )

        group.append(
            "csv_joiner",
            Joiner(
                "Match file and image metadata",
                allow_none=False,
                doc="""\
*(Used only if you want to extract metadata from the file and/or folder name
AND you're extracting metadata from a file)*

Match columns in your .csv file to image metadata items. If you are
using a CSV in conjunction with the filename/path metadata matching, you
might want to capture the metadata in common with both sources. For
example, you might be extracting the well tag from the image filename
while your CSV contains treatment dosage information paired with each
well. Therefore, you would want to let CellProfiler know that the well
tag extracted from the image filename and the well tag noted in the CSV
are in fact the one and the same.

This setting controls how rows in your CSV file are matched to different
images. Set the drop-downs to pair the metadata tags of the images and
the CSV, such that each row contains the corresponding tags. This can be
done for as many metadata correspondences as you may have for each
source; press |image0| to add more rows.

.. |image0| image:: {MODULE_ADD_BUTTON}
""".format(
                    **{"MODULE_ADD_BUTTON": image_resource("module_add.png")}
                ),
            ),
        )

        group.append(
            "wants_case_insensitive",
            Binary(
                "Use case insensitive matching?",
                False,
                doc="""\
*(Used only if "Match file and image metadata" is set)*

This setting controls whether row matching takes the metadata case into
account when matching.

-  Select "*{NO}*" so that metadata entries that only differ by case (for
   instance, “A01” and “a01”) will not match.
-  Select "*{YES}*" to match metadata entries that only differ by case.

|image0| If you note that your CSV metadata is
not being applied, your choice on this setting may be the culprit.

.. |image0| image:: {PROTIP_RECOMMEND_ICON}
""".format(
                    **{
                        "NO": "No",
                        "YES": "Yes",
                        "PROTIP_RECOMMEND_ICON": PROTIP_RECOMMEND_ICON,
                    }
                ),
            ),
        )

        group.imported_metadata_header_timestamp = 0
        group.imported_metadata_header_path = None
        group.imported_metadata_col_names = None
        group.imported_metadata_dicts_timestamp = 0
        group.imported_metadata_dicts_path = None
        group.imported_metadata_dicts = None
        # A temporary variable used to store a compiled regex object
        group.regex_pattern = None
        group.can_remove = can_remove
        if can_remove:
            group.append(
                "remover",
                RemoveSettingButton(
                    "", "Remove this extraction method", self.extraction_methods, group
                ),
            )
        self.extraction_methods.append(group)

    @staticmethod
    def csv_path(group):
        return os.path.join(
            group.csv_location.get_absolute_path(), group.csv_filename.value
        )

    def get_csv_header(self, group):
        """Get the header line from the imported extraction group's csv file"""
        csv_path = self.csv_path(group)
        if csv_path == group.imported_metadata_header_path:
            if group.csv_location.is_url():
                return group.imported_metadata_col_names
            if os.path.isfile(csv_path):
                timestamp = os.stat(csv_path).st_mtime
                if timestamp <= group.imported_metadata_header_timestamp:
                    return group.imported_metadata_col_names
        group.imported_metadata_header_timestamp = time.time()
        group.imported_metadata_header_path = csv_path
        try:
            if group.csv_location.is_url():
                url = generate_presigned_url(csv_path)
                fd = urllib.request.urlopen(url)
            else:
                fd = open(csv_path, "r")
            rdr = csv.DictReader(fd)
            group.imported_metadata_col_names = rdr.fieldnames
            fd.close()
        except Exception as e:
            LOGGER.error(f"Error while decoding CSV - {csv_path}: {e.strerror}")
            return None
        return group.imported_metadata_col_names

    def refresh_group_joiner(self, group):
        """Refresh the metadata entries for a group's joiner"""
        if group.extraction_method != X_IMPORTED_EXTRACTION:
            return
        #
        # Get the key set.
        #
        possible_keys = set(DEFAULT_METADATA_TAGS)
        for extract_group in self.extraction_methods:
            if extract_group is group:
                # Only want keys made before this group.
                break
            if extract_group.extraction_method == X_MANUAL_EXTRACTION:
                self.compile_regex(group)
                possible_keys.update(group.regex_pattern.groupindex.keys())
        joiner = group.csv_joiner
        assert isinstance(joiner, Joiner)
        joiner.entities[self.IPD_JOIN_NAME] = list(possible_keys)
        header_keys = self.get_csv_header(group)
        if header_keys is None:
            header_keys = ["None"]
        joiner.entities[self.CSV_JOIN_NAME] = header_keys

    @staticmethod
    def compile_regex(group):
        # Compile regex string into and object
        if group.source == XM_FILE_NAME:
            group.regex_pattern = re.compile(group.file_regexp.value)
        else:
            group.regex_pattern = re.compile(group.folder_regexp.value)

    def settings(self):
        result = [
            self.wants_metadata,
            self.data_type_choice,
            self.data_types,
            self.extraction_method_count,
        ]

        for group in self.extraction_methods:
            result += [
                group.extraction_method,
                group.source,
                group.file_regexp,
                group.folder_regexp,
                group.filter_choice,
                group.filter,
                group.csv_location,
                group.csv_joiner,
                group.wants_case_insensitive,
                group.csv_filename,
            ]
        return result

    def visible_settings(self):
        result = [self.wants_metadata]
        if self.wants_metadata:
            for group in self.extraction_methods:
                if group.can_remove:
                    result += [group.divider]
                result += [group.extraction_method]
                if group.extraction_method == X_MANUAL_EXTRACTION:
                    result += [group.source]
                    if group.source == XM_FILE_NAME:
                        result += [group.file_regexp]
                    elif group.source == XM_FOLDER_NAME:
                        result += [group.folder_regexp]
                    result += [group.filter_choice]
                    if group.filter_choice == F_FILTERED_IMAGES:
                        result += [group.filter]
                elif group.extraction_method == X_IMPORTED_EXTRACTION:
                    result += [
                        group.csv_location,
                        group.csv_filename,
                        group.filter_choice,
                    ]
                    if group.filter_choice == F_FILTERED_IMAGES:
                        result += [group.filter]
                    result += [group.csv_joiner, group.wants_case_insensitive]
                if group.can_remove:
                    result += [group.remover]
            result += [self.add_extraction_method_button]
            try:
                has_keys = len(self.get_dt_metadata_keys()) > 0
            except Exception as e:
                has_keys = False
            if has_keys:
                result += [self.dtc_divider, self.data_type_choice]
                if self.data_type_choice == DTC_CHOOSE:
                    result.append(self.data_types)
            result += [self.table]
        return result

    def example_file_fn(self):
        """Get an example file name for the regexp editor"""
        if self.filtered_file_list is not None and self.filtered_file_list:
            return self.filtered_file_list[0].filename
        return "PLATE_A01_s1_w11C78E18A-356E-48EC-B204-3F4379DC43AB.tif"

    def example_directory_fn(self):
        """Get an example directory name for the regexp editor"""
        if self.filtered_file_list is not None and self.filtered_file_list:
            return self.filtered_file_list[0].dirname
        return "/images/2012_01_12"

    def change_causes_prepare_run(self, setting):
        """Return True if changing the setting passed changes the image sets

        setting - the setting that was changed
        """
        return setting in self.settings()

    @classmethod
    def is_input_module(cls):
        return True

    def prepare_run(self, workspace):
        """Initialize the pipeline's metadata"""
        if workspace.pipeline.in_batch_mode():
            return True

        pipeline = workspace.pipeline
        assert isinstance(pipeline, Pipeline)
        filtered_file_list = pipeline.get_filtered_file_list(workspace)

        self.run_extraction(filtered_file_list)

        key_set = sorted(self.get_metadata_keys())
        pipeline.set_image_plane_details(None, key_set, self)
        return True

    def run_extraction(self, file_objects):
        if not self.wants_metadata.value:
            return
        # We first run through the methods and pre-prepare objects we'll use repeatedly.
        for group in self.extraction_methods:
            if group.extraction_method == X_MANUAL_EXTRACTION:
                # Compile regular expressions into proper objects. You can call regex.search directly with
                # the pattern string, but this wastes time converting it into a regex object during each call.
                self.compile_regex(group)
            elif group.extraction_method == X_IMPORTED_EXTRACTION:
                # Import the csv metadata as a list of dictionaries.
                self.import_csv_dict(group)
        for file_object in file_objects:
            file_object.clear_metadata()
            for group in self.extraction_methods:
                if group.filter_choice == F_FILTERED_IMAGES:
                    modpath = file_object.modpath
                    if not group.filter.evaluate((FileCollectionDisplay.NODE_FILE, modpath, None,)):
                        # Doesn't pass file filter for this extraction method
                        continue
                if group.extraction_method == X_MANUAL_EXTRACTION:
                    # REGEX Mode
                    if group.source == XM_FILE_NAME:
                        target = file_object.filename
                    else:
                        target = file_object.dirname
                    matches = group.regex_pattern.search(target)
                    if matches:
                        file_metadata = self.apply_data_types(matches.groupdict())
                        file_object.add_metadata(file_metadata)
                elif group.extraction_method == X_IMPORTED_EXTRACTION:
                    # Imported from CSV. Test the image against available dicts.
                    joins = group.csv_joiner.parse()
                    image_meta = file_object.metadata
                    valid = True
                    if group.imported_metadata_dicts is None:
                        valid = False
                    for candidate_dict in group.imported_metadata_dicts or []:
                        if None in candidate_dict:
                            # Extra columns without header labels were present. Delete them.
                            del candidate_dict[None]
                        candidate_dict = self.apply_data_types(candidate_dict)
                        valid = True
                        for join_dict in joins:
                            csv_key = join_dict[CSV_JOIN_NAME]
                            image_key = join_dict[IPD_JOIN_NAME]
                            csv_val = candidate_dict[csv_key]
                            image_val = self.apply_data_type(image_meta.get(image_key, ""), self.get_data_type(image_key))
                            if group.wants_case_insensitive:
                                if isinstance(csv_val, str):
                                    csv_val = csv_val.lower()
                                if isinstance(image_val, str):
                                    image_val = image_val.lower()
                            if csv_val != image_val:
                                valid = False
                                break
                        if valid:
                            file_object.add_metadata(candidate_dict)
                            break
                    if not valid:
                        LOGGER.info(f"No matching metadata found for {file_object.filename}")
                        break
                    pass
                else:
                    raise NotImplementedError(f"Invalid extraction method '{group.extraction_method}'")
            if FTR_WELL not in file_object.metadata:
                row_key = next((x for x in ROW_KEYS if x in file_object.metadata), None)
                col_key = next((x for x in COL_KEYS if x in file_object.metadata), None)
                if row_key is not None and col_key is not None:
                    col = file_object.metadata[col_key]
                    row = file_object.metadata[row_key]
                    if row.isdigit():
                        row = int(row)
                    if isinstance(row, int):
                        if 0 < row <= 24:
                            row = string.ascii_uppercase[row - 1]
                    well = row + col
                    file_object.set_metadata(FTR_WELL, well)
        for group in self.extraction_methods:
            if group.extraction_method == X_IMPORTED_EXTRACTION:
                # Extraction is done, clear the dict list to save memory.
                del group.imported_metadata_dicts
                group.imported_metadata_dicts = None
                group.imported_metadata_dicts_timestamp = 0
                group.imported_metadata_dicts_path = None

    def import_csv_dict(self, group):
        csv_path = self.csv_path(group)
        if csv_path == group.imported_metadata_dicts_path and group.imported_metadata_dicts is not None:
            if group.csv_location.is_url():
                return group.imported_metadata_dicts
            if os.path.isfile(csv_path):
                timestamp = os.stat(csv_path).st_mtime
                if timestamp <= group.imported_metadata_dicts_timestamp:
                    return group.imported_metadata_dicts
        group.imported_metadata_dicts_timestamp = time.time()
        group.imported_metadata_dicts_path = csv_path
        try:
            if group.csv_location.is_url():
                url = generate_presigned_url(csv_path)
                fd = urllib.request.urlopen(url)
            else:
                fd = open(csv_path, "r")
            rdr = csv.DictReader(fd)
            group.imported_metadata_dicts = [x for x in rdr]
            fd.close()
        except Exception as e:
            print("Error while decoding CSV:", e)
            return None

    def run(self, workspace):
        pass
        
    def on_activated(self, workspace):
        self.workspace = workspace
        self.pipeline = workspace.pipeline
        self.filtered_file_list = workspace.pipeline.get_filtered_file_list(workspace)
        if self.pipeline.file_list_edited:
            self.pipeline.file_list_edited = False
        for group in self.extraction_methods:
            if group.extraction_method == X_IMPORTED_EXTRACTION:
                self.refresh_group_joiner(group)
        self.table.clear_rows()
        self.table.clear_columns()
        if not self.pipeline.file_list_edited:
            self.update_table()
        else:
            # File list has changed, reset 'needs metadata extraction' flag
            pass

    def on_setting_changed(self, setting, pipeline):
        """Update the imported extraction joiners on setting changes"""
        if not self.wants_metadata:
            print("Clearing metadata")
            for file_object in self.filtered_file_list:
                file_object.clear_metadata()
            return
        visible_settings = self.visible_settings()
        if setting == self.data_types or setting == self.data_type_choice:
            # The data types affect the joiner's matching
            setting_idx = len(self.visible_settings())
        elif setting in visible_settings:
            setting_idx = visible_settings.index(setting)
        else:
            setting_idx = len(self.visible_settings())
        for group in self.extraction_methods:
            if group.extraction_method == X_IMPORTED_EXTRACTION:
                idx = max(
                    *list(
                        map(
                            visible_settings.index,
                            [
                                group.csv_joiner,
                                group.csv_location,
                                group.wants_case_insensitive,
                            ],
                        )
                    )
                )
                if idx < setting_idx:
                    continue
                self.refresh_group_joiner(group)

    def update_table(self, *args):
        self.run_extraction(self.filtered_file_list)
        columns = set(self.get_metadata_keys())
        columns.discard(COL_SERIES)
        columns.discard(COL_INDEX)
        columns = [COL_PATH, COL_SERIES, COL_INDEX] + sorted(list(columns))
        self.table.clear_columns()
        self.table.clear_rows()
        data = []
        for file_object in self.filtered_file_list:
            md_keys = file_object.metadata
            row = [md_keys.get(k, None) for k in columns]
            data.append(row)

        for i, column in enumerate(columns):
            self.table.insert_column(i, column)

        self.table.add_rows(columns, data)
        report_progress("MetadataCount", None, "Found %d rows" % len(data))

    def on_deactivated(self):
        self.pipeline = None
        self.filtered_file_list = None

    def prepare_to_create_batch(self, workspace, fn_alter_path):
        """Alter internal paths for batch creation"""
        for group in self.extraction_methods:
            if group.extraction_method == X_IMPORTED_EXTRACTION:
                group.csv_location.alter_for_create_batch(fn_alter_path)

    def prepare_settings(self, setting_values):
        """Prepare the module to receive the settings"""
        #
        # Set the number of extraction methods based on the extraction method
        # count.
        #
        n_extraction_methods = int(setting_values[IDX_EXTRACTION_METHOD_COUNT])
        if len(self.extraction_methods) > n_extraction_methods:
            del self.extraction_methods[n_extraction_methods:]

        while len(self.extraction_methods) < n_extraction_methods:
            self.add_extraction_method()

    def validate_module(self, pipeline):
        """Validate the module settings

        pipeline - current pipeline

        Metadata throws an exception if any of the metadata tags collide with
        tags that can be automatically extracted.
        """
        for group in self.extraction_methods:
            if group.extraction_method == X_MANUAL_EXTRACTION:
                re_setting = (
                    group.file_regexp
                    if group.source == XM_FILE_NAME
                    else group.folder_regexp
                )
                for token in find_metadata_tokens(re_setting.value):
                    if token.upper() in [
                        reservedtag.upper() for reservedtag in RESERVED_METADATA_KEYS
                    ]:
                        raise ValidationError(
                            'The metadata tag, "%s", is reserved for use by CellProfiler.'
                            " Please use some other tag name." % token,
                            re_setting,
                        )

    def get_metadata_keys(self):
        """Return a collection of metadata keys to be associated with files"""
        keys = set([COL_URL])
        if not self.wants_metadata:
            return keys
        keys = set(DEFAULT_METADATA_TAGS)
        for extract_group in self.extraction_methods:
            if extract_group.extraction_method == X_MANUAL_EXTRACTION:
                self.compile_regex(extract_group)
                keys.update(extract_group.regex_pattern.groupindex.keys())
            elif extract_group.extraction_method == X_IMPORTED_EXTRACTION:
                keys.update(self.get_csv_header(extract_group) or '')
        if FTR_WELL not in keys and any(k in keys for k in ROW_KEYS) and any(k in keys for k in COL_KEYS):
            keys.add(FTR_WELL)
        return keys

    def get_dt_metadata_keys(self):
        """Get the metadata keys which can have flexible datatyping

        """
        return list(
            filter(
                (lambda k: k not in self.NUMERIC_DATA_TYPES), self.get_metadata_keys() or []
            )
        )

    NUMERIC_DATA_TYPES = (
        MD_SIZE_C,
        MD_SIZE_T,
        MD_SIZE_Z,
        MD_SIZE_X,
        MD_SIZE_Y,
        C_SERIES,
        C_FRAME,
        C_C,
        C_T,
        C_Z,
    )

    def get_data_type(self, key):
        """Get the data type for a particular metadata key"""
        if isinstance(key, str):
            return self.get_data_type([key]).get(key, COLTYPE_VARCHAR)
        result = {}
        if self.data_type_choice == DTC_CHOOSE:
            data_types = DataTypes.decode_data_types(self.data_types.value_text)
        for k in key:
            if k in self.NUMERIC_DATA_TYPES:
                result[k] = COLTYPE_INTEGER
            elif self.data_type_choice == DTC_CHOOSE:
                dt = data_types.get(k, DataTypes.DT_TEXT)
                if dt == DataTypes.DT_TEXT:
                    result[k] = COLTYPE_VARCHAR
                elif dt == DataTypes.DT_INTEGER:
                    result[k] = COLTYPE_INTEGER
                elif dt == DataTypes.DT_FLOAT:
                    result[k] = COLTYPE_FLOAT
            else:
                result[k] = COLTYPE_VARCHAR

        return result

    def apply_data_types(self, input_dict):
        return {k: self.apply_data_type(v, self.get_data_type(k)) for k, v, in input_dict.items()}

    @staticmethod
    def apply_data_type(value, data_type):
        if data_type == DataTypes.DT_TEXT:
            return str(value)
        elif data_type == DataTypes.DT_INTEGER:
            try:
                return int(value)
            except ValueError:
                LOGGER.warning(f"Metadata value {value} cannot be interpreted as an integer number.")
                return value
        elif data_type == DataTypes.DT_FLOAT:
            try:
                return float(value)
            except ValueError:
                LOGGER.warning(f"Metadata value {value} cannot be interpreted as a floating point number.")
                return value

        return value

    def wants_case_insensitive_matching(self, key):
        """Return True if the key should be matched using case-insensitive matching

        key - key to check.

        Currently, there is a case-insensitive matching flag in the
        imported metadata matcher. Perhaps this should be migrated into
        the data types control, but for now, we look for the key to be
        present in the joiner for any imported metadata matcher.
        """
        if not self.wants_metadata:
            return False
        for group in self.extraction_methods:
            if (
                    group.extraction_method == X_IMPORTED_EXTRACTION
                    and group.wants_case_insensitive
            ):
                joins = group.csv_joiner.parse()
                for join in joins:
                    if key in list(join.values()):
                        return True
        return False

    def get_measurement_columns(self, pipeline):
        """Get the metadata measurements collected by this module"""
        key_types = pipeline.get_available_metadata_keys()
        result = []
        for key, coltype in list(key_types.items()):
            if self.data_type_choice == DTC_CHOOSE:
                data_type = self.get_data_type(key)
                if data_type == DataTypes.DT_NONE:
                    continue
                elif data_type == DataTypes.DT_INTEGER:
                    data_type = COLTYPE_INTEGER
                elif data_type == DataTypes.DT_FLOAT:
                    data_type = COLTYPE_FLOAT
                else:
                    data_type = COLTYPE_VARCHAR_FILE_NAME
            else:
                data_type = COLTYPE_VARCHAR_FILE_NAME
            result.append(("Image", "_".join((C_METADATA, key)), data_type,))
        return result

    def get_categories(self, pipeline, object_name):
        """Return the measurement categories for a particular object"""
        if object_name == "Image" and len(self.get_metadata_keys()) > 0:
            return [C_METADATA]
        return []

    def get_measurements(self, pipeline, object_name, category):
        if object_name == "Image" and category == C_METADATA:
            keys = self.get_metadata_keys()
            return keys
        return []

    def upgrade_settings(self, setting_values, variable_revision_number, module_name):
        if variable_revision_number == 1:
            n_groups = int(setting_values[IDX_EXTRACTION_METHOD_COUNT_V1])
            new_setting_values = setting_values[:IDX_EXTRACTION_METHOD_V1]
            for i in range(n_groups):
                new_setting_values += setting_values[
                                      (IDX_EXTRACTION_METHOD_V1 + LEN_EXTRACTION_METHOD_V1 * i): (
                                              IDX_EXTRACTION_METHOD_V1 + LEN_EXTRACTION_METHOD_V1 * (i + 1)
                                      )
                                      ]
                new_setting_values.append("No")
            setting_values = new_setting_values
            variable_revision_number = 2

        if variable_revision_number == 2:
            # Changed naming of extraction methods, metadata sources and filtering choices
            n_groups = int(setting_values[IDX_EXTRACTION_METHOD_COUNT_V2])
            new_setting_values = setting_values[:IDX_EXTRACTION_METHOD_V2]
            for i in range(n_groups):
                group = setting_values[
                        (IDX_EXTRACTION_METHOD_V2 + LEN_EXTRACTION_METHOD * i): (
                                IDX_EXTRACTION_METHOD_V2 + LEN_EXTRACTION_METHOD * (i + 1)
                        )
                        ]
                group[0] = (
                    X_AUTOMATIC_EXTRACTION
                    if group[0] == "Automatic"
                    else (
                        X_MANUAL_EXTRACTION
                        if group[0] == "Manual"
                        else X_IMPORTED_EXTRACTION
                    )
                )
                group[1] = (
                    XM_FILE_NAME if group[1] == "From file name" else XM_FOLDER_NAME
                )
                group[4] = (
                    F_FILTERED_IMAGES
                    if group[4] == "Images selected using a filter"
                    else F_ALL_IMAGES
                )
                new_setting_values += group
            setting_values = new_setting_values
            variable_revision_number = 3

        if variable_revision_number == 3:
            # Added data types
            setting_values = (
                    setting_values[:IDX_EXTRACTION_METHOD_COUNT_V3]
                    + [DTC_TEXT, "{}"]
                    + setting_values[IDX_EXTRACTION_METHOD_COUNT_V3:]
            )
            variable_revision_number = 4

        if variable_revision_number == 4:
            # Allow metadata CSVs to be loaded from default io directories.
            groups = []
            n_groups = int(setting_values[3])
            for group_idx in range(n_groups):
                # group offset: 4
                # no. group settings: 9
                group = setting_values[4 + (group_idx * 9): 4 + ((group_idx + 1) * 9)]

                csv_location = group[6]
                directory, filename = os.path.split(csv_location)
                if any(
                        [
                            csv_location.lower().startswith(scheme)
                            for scheme in ["http:", "https:", "ftp:"]
                        ]
                ):
                    directory_choice = URL_FOLDER_NAME
                else:
                    directory_choice = ABSOLUTE_FOLDER_NAME

                group[6] = "{}|{}".format(directory_choice, directory)
                group += [filename]
                groups += [group]

            setting_values[4:] = sum(groups, [])
            variable_revision_number = 5
        if variable_revision_number == 5:
            # Add record of group metadata storage.
            new_setting_values = setting_values
            groups = []
            n_groups = int(setting_values[3])
            for group_idx in range(n_groups):
                # group offset: 4
                # no. group settings: 10
                group = setting_values[
                        4 + (group_idx * 10): 4 + ((group_idx + 1) * 10)
                        ]
                group += ["No"]
                groups += [group]
            new_setting_values[4:] = sum(groups, [])
            setting_values = new_setting_values
            variable_revision_number = 6
        if variable_revision_number == 6:
            # Remove that record of group metadata storage. Remove file header scrapers.
            new_setting_values = setting_values[:4]
            n_groups = int(setting_values[3])
            new_n_groups = 0
            for group_idx in range(n_groups):
                # group offset: 4
                # no. group settings: 11
                group = setting_values[
                        4 + (group_idx * 11): 4 + ((group_idx + 1) * 11)
                        ]
                if group[0] == X_AUTOMATIC_EXTRACTION:
                    self.removed_automatic_extraction = True
                    continue
                new_setting_values += group[:-1]
                new_n_groups += 1
            new_setting_values[3] = str(new_n_groups)
            setting_values = new_setting_values
            variable_revision_number = 7

        return setting_values, variable_revision_number

    def volumetric(self):
        return True



import logging

from .._setting import Setting
from .._validation_error import ValidationError


LOGGER = logging.getLogger(__name__)

class Range(Setting):
    """A setting representing a range between two values"""

    valid_format_text = '"%s" is formatted incorrectly'

    def __init__(self, text, value, minval=None, maxval=None, *args, **kwargs):
        """Initialize a range

        text - helpful text to be displayed to the user

        value - default value as a string, should be in the form <min>,<max>

        minval - the minimum value for the range (or None if none)

        maxval - the maximum value of the range (or None if none)
        """
        super(Range, self).__init__(text, value, *args, **kwargs)
        self._minval = minval
        self._maxval = maxval
        self.__default_min = self.min
        self.__default_max = self.max

    def str_to_value(self, value_str):
        """Convert a min/max value as a string to the native type"""
        raise NotImplementedError("str_to_value must be implemented in derived class")

    def value_to_str(self, value):
        """Convert a string to a min/max value in the native type"""
        raise NotImplementedError("value_to_str must be implemented in derived class")

    def get_value(self):
        """Return the value of this range as a min/max tuple"""
        return self.min, self.max

    def set_value(self, value):
        """Set the value of this range using either a string or a two-tuple"""
        if isinstance(value, str):
            self.set_value_text(value)
        elif hasattr(value, "__getitem__") and len(value) == 2:
            self.set_value_text(",".join([self.value_to_str(v) for v in value]))
        else:
            raise ValueError("Value for range must be a string or two-tuple")

    def get_min_text(self):
        """Get the minimum of the range as a text value"""
        return self.get_value_text().split(",")[0]

    def get_min(self):
        """Get the minimum of the range as a number"""
        try:
            value = self.str_to_value(self.get_min_text())
            if self._minval is not None and value < self._minval:
                return self._minval
            return value
        except:
            return self.__default_min

    def get_max_text(self):
        """Get the maximum of the range as a text value"""
        vv = self.get_value_text().split(",")
        if len(vv) < 2:
            return ""
        return vv[1]

    def get_max(self):
        """Get the maximum of the range as a number"""
        try:
            value = self.str_to_value(self.get_max_text())
            if self._maxval is not None and value > self._maxval:
                return self._maxval
            return value
        except:
            return self.__default_max

    def compose_min_text(self, value):
        """Return the text value that would set the minimum to the proposed value

        value - the proposed minimum value as text
        """
        return ",".join((value, self.get_max_text()))

    def set_min(self, value):
        """Set the minimum part of the value, given the minimum as a #"""
        self.set_value_text(self.compose_min_text(self.value_to_str(value)))

    def compose_max_text(self, value):
        """Return the text value that would set the maximum to the proposed value

        value - the proposed maximum value as text
        """
        return ",".join((self.get_min_text(), value))

    def set_max(self, value):
        """Set the maximum part of the value, given the maximum as a #"""
        self.set_value_text(self.compose_max_text(self.value_to_str(value)))

    min = property(get_min, set_min)
    min_text = property(get_min_text)
    max = property(get_max, set_max)
    max_text = property(get_max_text)

    def set_value_text(self, value):
        super(Range, self).set_value_text(value)
        try:
            self.test_valid(None)
            self.__default_min = self.min
            self.__default_max = self.max
        except:
            LOGGER.debug("Illegal value in range setting: %s" % value)

    def test_valid(self, pipeline):
        values = self.value_text.split(",")
        if len(values) < 2:
            raise ValidationError(
                "Minimum and maximum values must be separated by a comma", self
            )
        if len(values) > 2:
            raise ValidationError("Only two values allowed", self)
        for value in values:
            try:
                self.str_to_value(value)
            except:
                raise ValidationError(self.valid_format_text % value, self)
        v_min, v_max = [self.str_to_value(value) for value in values]
        if self._minval is not None and self._minval > v_min:
            raise ValidationError(
                "%s can't be less than %s"
                % (self.min_text, self.value_to_str(self._minval)),
                self,
            )
        if self._maxval is not None and self._maxval < v_max:
            raise ValidationError(
                "%s can't be greater than %s"
                % (self.max_text, self.value_to_str(self._maxval)),
                self,
            )
        if v_min > v_max:
            raise ValidationError(
                "%s is greater than %s" % (self.min_text, self.max_text), self
            )

    def eq(self, x):
        """If the operand is a sequence, true if it matches the min and max"""
        if hasattr(x, "__getitem__") and len(x) == 2:
            return x[0] == self.min and x[1] == self.max
        return False



"""
Run pipelines on imagesets to produce measurements.
"""

import sys


use_analysis = True

DEBUG = "DEBUG"
ANNOUNCE_DONE = "DONE"


###############################
# Request, Replies, Events
###############################


if sys.platform == "darwin":
    pass

if __name__ == "__main__":
    # This is an ugly hack, but it's necesary to unify the Request/Reply
    # classes above, so that regardless of whether this is the current module,
    # or a separately imported one, they see the same classes.
    import cellprofiler_core.analysis
    from ._runner import Runner

    globals().update(cellprofiler_core.analysis.__dict__)

    Runner.start_workers(2)
    Runner.stop_workers()



from ._event import Event


class URLsRemoved(Event):
    def __init__(self, urls):
        super(self.__class__, self).__init__()
        self.urls = urls

    def event_type(self):
        return "URLs removed from file list"



from ._monochrome_image import MonochromeImage


class MaskImage(MonochromeImage):
    """Provide a boolean image, converting nonzero to True, zero to False if needed"""

    def __init__(self,
                 name,
                 url,
                 series,
                 index,
                 channel,
                 volume=False,
                 spacing=None,
                 z=None,
                 t=None
                ):
        MonochromeImage.__init__(
            self,
            name,
            url,
            rescale_range=None,
            metadata_rescale=False,
            series=series,
            index=index,
            channel=channel,
            volume=volume,
            spacing=spacing,
            z=z,
            t=t
        )

    def provide_image(self, image_set):
        image = MonochromeImage.provide_image(self, image_set)
        if image.pixel_data.dtype.kind != "b":
            image.pixel_data = image.pixel_data != 0
        return image



import io
from urllib.parse import urlparse
import logging

import imageio
from cellprofiler_core.readers.imageio_reader import ImageIOReader
from cellprofiler_core.readers.imageio_reader import SUPPORTED_EXTENSIONS as IMAGEIO_SUPPORTED_EXTENSIONS
from cellprofiler_core.readers.imageio_reader import SEMI_SUPPORTED_EXTENSIONS as IMAGEIO_SEMI_SUPPORTED_EXTENSIONS
from google.cloud import storage
from google.auth.exceptions import DefaultCredentialsError
from google.cloud.exceptions import NotFound

LOGGER = logging.getLogger(__name__)

SUPPORTED_EXTENSIONS = {'.tiff'}
SEMI_SUPPORTED_EXTENSIONS = IMAGEIO_SUPPORTED_EXTENSIONS.union(IMAGEIO_SEMI_SUPPORTED_EXTENSIONS).difference(SUPPORTED_EXTENSIONS)
SUPPORTED_SCHEMES = {'gs'}


class GcsReader(ImageIOReader):
    """
    Reads images from bucket(s) in Google Cloud Storage via user's Application
    Default Credential (ADC).
    
    Once the file is downloaded from GCS, this reader leverages ImageIO Reader's
    image loading methods.

    Prerequisites:
    User has authenticated with Google Cloud Storage Application Default Credential
    (ADC) by running the command: "gcloud auth application-default login",
    or is running CellProfiler in an environment where this credential has already
    been configured for them such as app.terra.bio or CloudShell.
    """

    reader_name = "Google Cloud Storage"
    variable_revision_number = 1
    supported_filetypes = SUPPORTED_EXTENSIONS.union(SUPPORTED_EXTENSIONS)
    supported_schemes = SUPPORTED_SCHEMES

    @classmethod
    def supports_format(cls, image_file, allow_open=False, volume=False):
        """This function needs to evaluate whether a given ImageFile object
        can be read by this reader class.

        Return value should be an integer representing suitability:
        -1 - 'I can't read this at all'
        1 - 'I am the one true reader for this format, don't even bother checking any others'
        2 - 'I am well-suited to this format'
        3 - 'I can read this format, but I might not be the best',
        4 - 'I can give it a go, if you must'

        The allow_open parameter dictates whether the reader is permitted to read the file when
        making this decision. If False the decision should be made using file extension only.
        Any opened files should be closed before returning.

        The volume parameter specifies whether the reader will need to return a 3D array.
        """
        if image_file.scheme not in SUPPORTED_SCHEMES:
            return -1

        if image_file.file_extension in SUPPORTED_EXTENSIONS:
            return 1

        if image_file.file_extension in SEMI_SUPPORTED_EXTENSIONS:
            return 3

        return 4

    def get_reader(self, volume=False):
        image_resource = self.__download_blob(self.file.url)
        if self._reader is None or volume != self._volume:
            if volume:
                self._reader = imageio.get_reader(image_resource, mode='v')
            else:
                self._reader = imageio.get_reader(image_resource, mode='i')
            self._volume = volume
        return self._reader

    @classmethod
    def supports_url(cls):
        return True

    @staticmethod
    def get_settings():
        return []

    def __decode_gcs_url(self, url):
        p = urlparse(url)
        file_path = p.path[1:] if p.path.startswith("/") else p.path
        bucket_id = p.netloc
        return bucket_id, file_path
  
    def __download_blob(self, url):
        if url:
            # Create client to access Google Cloud Storage.
            try:
                client = storage.Client()
                # Get bucket name, file path from URL.
                bucket_id, file_path = self.__decode_gcs_url(url)
                # Get bucket.
                bucket = client.bucket(bucket_id)
                # Download blob object to local file path.
                blob = bucket.blob(file_path)
                blob_bytes = io.BytesIO(blob.download_as_bytes())
                return blob_bytes
            except DefaultCredentialsError as e:
               LOGGER.error(e.args[0])
               return url
            except NotFound as e:
                # 404 error
                LOGGER.error(e)
                return url



from cellprofiler_core.preferences import (
    ABSOLUTE_FOLDER_NAME,
    DEFAULT_INPUT_SUBFOLDER_NAME,
    DEFAULT_OUTPUT_SUBFOLDER_NAME,
)
from cellprofiler_core.utilities.image import image_resource

DEFINITION_OBJECT = """\
In CellProfiler, we use the term *object* as a generic term to refer to
an identified feature in an image, usually an organism, cell, or cellular
compartment (for example, nuclei, cells, colonies, worms).
"""
REGEXP_HELP_REF = """\
**Details on regular expressions:**

A *regular expression* is a general term referring to a method of
searching for pattern matches in text. There is a high learning curve to
using them, but are quite powerful once you understand the basics.

Patterns are specified using combinations of metacharacters and literal
characters. There are a few classes of metacharacters, partially listed
below. Some helpful links follow:

-  A more extensive explanation of regular expressions can be found
   `here`_
-  A helpful quick reference can be found
   `here <http://www.cheatography.com/davechild/cheat-sheets/regular-expressions/>`__
-  `Pythex`_ provides quick way to test your regular expressions. Here
   is an `example`_ to capture information from a common microscope
   nomenclature.

.. _here: http://docs.python.org/2/howto/regex.html
.. _Pythex: http://pythex.org/
.. _example: http://pythex.org/?regex=Channel%5B1-2%5D-%5B0-9%5D%7B2%7D-(%3FP%3CWellRow%3E%5BA-H%5D)-(%3FP%3CWellColumn%3E%5B0-9%5D%7B2%7D)%5C.tif&test_string=Channel1-01-A-01.tif&ignorecase=0&multiline=0&dotall=0&verbose=0
"""
FILTER_RULES_BUTTONS_HELP = """\
Clicking the rule menus shows you all the file *attributes*, *operators*
and *conditions* you can specify to narrow down the image list.

#. For each rule, first select the *attribute* that the rule is to be
   based on. For example, you can select “File” to define a rule that
   will filter files on the basis of their filename.
#. The *operator* drop-down is then updated with operators applicable to
   the attribute you selected. For example, if you select “File” as the
   attribute, the operator menu includes text operators such as
   *Contain* or *Starts with*. On the other hand, if you select
   “Extension” as the attribute, you can choose the logical operators
   “Is” or “Is not” from the menu.
#. In the operator drop-down menu, select the operator you want to use.
   For example, if you want to match data exactly, you may want the
   “Exactly match” or the “Is” operator. If you want the condition to be
   more loose, select an operator such as “Contains”.
#. Use the *condition* box to type the condition you want to match. The
   more you type, the more specific the condition is.

   -  As an example, if you create a new filter and select *File* as the
      attribute, then select “Does” and “Contain” as the operators, and
      type “BBBC013” as the condition, the filter finds all files that
      include the text “BBBC013”, such as “BBBC013-1.tif” “BBBC013-2.jpg”,
      “1-BBBC013-A01.BMP” and so on.
   -  If you select “Does” and “Start with” as the operators and
      “BBBC013-1” in the Condition box, the rule will includes such files
      as “BBBC013-1.tif” “BBBC013-1-A01.png”, and so on.

.. image:: {IMAGES_USING_RULES_ICON}
   :width: 100%

You can also create regular expressions (an advanced syntax for
pattern matching) in order to select particular files.

To add another rule, click the plus buttons to the right of each rule.
Subtract an existing rule by clicking the minus button.

You can also link a set of rules by choosing the logical expression
*All* or *Any*. If you use *All* logical expression, all the rules must
be true for a file to be included in the File list. If you use the *Any*
option, only one of the conditions has to be met for a file to be
included.

If you want to create more complex rules (e.g, some criteria matching
all rules and others matching any), you can create sets of rules, by
clicking the ellipsis button (to the right of the plus button). Repeat
the above steps to add more rules to the filter until you have all the
conditions you want to include.

{REGEXP_HELP_REF}
""".format(
    **{
        "IMAGES_USING_RULES_ICON": image_resource("Images_UsingRules.png"),
        "REGEXP_HELP_REF": REGEXP_HELP_REF,
    }
)
HELP_ON_MEASURING_DISTANCES = """\
To measure distances in an open image, use the “Measure length” tool
under *Tools* in the display window menu bar. If you click on an image
and drag, a line will appear between the two endpoints, and the distance
between them will be shown at the right-most portion of the bottom panel.\
"""
HELP_ON_MEASURING_INTENSITIES = """\
Note that for publication purposes, the units of intensity from
microscopy images are usually described as “Intensity units” or
“Arbitrary intensity units” because microscopes are not calibrated to an
absolute scale. Also, it is important to note whether you are reporting
the mean vs. the integrated intensity, so specify “Mean intensity
units” or “Integrated intensity units” accordingly.

Keep in mind that the default behavior in CellProfiler is to rescale the
image intensity from 0 to 1 by dividing all pixels in the image by the
maximum possible intensity value. This “maximum possible” value is
defined by the “Set intensity range from” setting in **NamesAndTypes**;
see the help for that setting for more details.
"""
HELP_ON_PIXEL_INTENSITIES = """\
To view pixel intensities in an open image, use the pixel intensity tool
which is available in any open display window. When you move your mouse
over the image, the pixel intensities will appear in the bottom bar of
the display window.\
"""
IO_FOLDER_CHOICE_HELP_TEXT = """\
You can choose among the following options which are common to all file
input/output modules:

-  *Default Input Folder*: Use the default input folder.
-  *Default Output Folder:* Use the default output folder.
-  *Elsewhere…*: Use a particular folder you specify.
-  *Default input directory sub-folder*: Enter the name of a subfolder
   of the default input folder or a path that starts from the default
   input folder.
-  *Default output directory sub-folder*: Enter the name of a subfolder
   of the default output folder or a path that starts from the default
   output folder.

*Elsewhere* and the two sub-folder options all require you to enter an
additional path name. You can use an *absolute path* (such as
“C:\\\\imagedir\\\\image.tif” on a PC) or a *relative path* to specify
the file location relative to a directory, which makes the pipeline
more flexible for future runs):

-  Use one period to represent the current directory. For example, if
   you choose *Default Input Folder sub-folder*, you can enter
   “./MyFiles” to look in a folder called “MyFiles” that is contained
   within the Default Input Folder.
-  Use two periods “..” to move up one folder level. For example, if you
   choose *Default Input Folder sub-folder*, you can enter “../MyFolder”
   to look in a folder called “MyFolder” at the same level as the
   Default Input Folder.\
"""
USING_METADATA_GROUPING_HELP_REF = """\
Please see the **Groups** module for more details on the proper use of
metadata for grouping.
"""
USING_METADATA_HELP_REF = """\
Please see the **Metadata** module for more details on metadata
collection and usage.
"""
USING_METADATA_TAGS_REF = """\
You can insert a previously defined metadata tag by either using:

-  The insert key
-  A right mouse button click inside the control
-  In Windows, the Context menu key, which is between the Windows key
   and Ctrl key

The inserted metadata tag will appear in green. To change a previously
inserted metadata tag, navigate the cursor to just before the tag and
either:

-  Use the up and down arrows to cycle through possible values.
-  Right-click on the tag to display and select the available values.
"""
IO_WITH_METADATA_HELP_TEXT = """\
For *{ABSOLUTE_FOLDER_NAME}*, *{DEFAULT_INPUT_SUBFOLDER_NAME}* and
*{DEFAULT_OUTPUT_SUBFOLDER_NAME}*, if you have metadata associated
with your images via **Metadata** module, you can name the folder using any
metadata tags for which all images in each individual image set have the same value.

-  Example: if you had extracted "*Plate*", "*Well*", and "*Channel*" metadata
   from your images, for most pipelines folders based on "*Plate*" or "*Well*" would work since
   each individual image set would come only from a single well on a single plate, but
   folders based on "*Channel*" would not work as each individual image set might
   contain many channels.

{USING_METADATA_TAGS_REF}

For instance, if you have a metadata tag named “Plate”, you can create a
per-plate folder by selecting one of the subfolder options and then
specifying the subfolder name as “\\g<Plate>”. The module will
substitute the metadata values for the current image set for any
metadata tags in the folder name.

{USING_METADATA_HELP_REF}
""".format(
    **{
        "ABSOLUTE_FOLDER_NAME": ABSOLUTE_FOLDER_NAME,
        "DEFAULT_INPUT_SUBFOLDER_NAME": DEFAULT_INPUT_SUBFOLDER_NAME,
        "DEFAULT_OUTPUT_SUBFOLDER_NAME": DEFAULT_OUTPUT_SUBFOLDER_NAME,
        "USING_METADATA_HELP_REF": USING_METADATA_HELP_REF,
        "USING_METADATA_TAGS_REF": USING_METADATA_TAGS_REF,
    }
)
HELP_ON_SAVING_OBJECTS = """\
*Note on saving images:* You can pass the objects along to the
*Object Processing* module **ConvertObjectsToImage** to create an image.
This image can be saved with the **SaveImages** module. Additionally,
you can use the **OverlayOutlines** or **OverlayObjects** module to
overlay outlines or objects, respectively, on a base image.
The resulting image can also be saved with the **SaveImages** module.
"""
StrelImage = image_resource("structuringelement.png")
HELP_FOR_STREL = """\
The structuring element is the shape that will be applied in any morphological
operation. The structuring element is centered on each pixel and the shape and size
selected will determine what neighborhood around that pixel will be affected by
that operation. See image below for an example of dilating a starting square of 11
pixel diameter with various structuring elements.

|StrelImage|

.. |StrelImage| image:: {StrelImage}
""".format(
    **{"StrelImage": StrelImage}
)
PROTIP_RECOMMEND_ICON = image_resource("thumb-up.png")
PROTIP_AVOID_ICON = image_resource("thumb-down.png")
TECH_NOTE_ICON = image_resource("gear.png")



class Event:
    """Something that happened to the pipeline and was indicated to the listeners
    """

    def __init__(self, is_pipeline_modification=False, is_image_set_modification=False):
        self.is_pipeline_modification = is_pipeline_modification
        self.is_image_set_modification = is_image_set_modification

    def event_type(self):
        raise NotImplementedError(
            "AbstractPipelineEvent does not implement an event type"
        )


class CancelledException(Exception):
    """Exception issued by the analysis worker indicating cancellation by UI

    This is here in order to solve some import dependency problems
    """

    pass


class PipelineLoadCancelledException(Exception):
    """Exception thrown if user cancels pipeline load"""

    pass



import os
import re
import sys
import tempfile

from .hdf5_dict import get_top_level_group
from .hdf5_dict import VERSION


def get_length_from_varchar(x):
    """Retrieve the length of a varchar column from its coltype def"""
    m = re.match(r"^varchar\(([0-9]+)\)$", x)
    if m is None:
        return None
    return int(m.groups()[0])


def make_temporary_file():
    """Make a temporary file to use for backing measurements data

    returns a file descriptor (that should be closed when done) and a
    file name.
    """
    from ..preferences import get_temporary_directory

    temporary_directory = get_temporary_directory()
    if not (
        os.path.exists(temporary_directory) and os.access(temporary_directory, os.W_OK)
    ):
        temporary_directory = None
    return tempfile.mkstemp(
        prefix="Cpmeasurements", suffix=".hdf5", dir=temporary_directory
    )


def load_measurements(filename, dest_file=None, run_name=None, image_numbers=None):
    """Load measurements from an HDF5 file

    filename - path to file containing the measurements or file-like object
               if .mat

    dest_file - path to file to be created. This file is used as the backing
                store for the measurements.

    run_name - name of the run (an HDF file can contain measurements
               from multiple runs). By default, takes the last.

    returns a Measurements object
    """
    from ..measurement import Measurements

    HDF5_HEADER = (
        chr(137) + chr(72) + chr(68) + chr(70) + chr(13) + chr(10) + chr(26) + chr(10)
    )
    if hasattr(filename, "seek"):
        filename.seek(0)
        header = filename.read(len(HDF5_HEADER))
        filename.seek(0)
    else:
        fd = open(filename, "rb")
        header = fd.read(len(HDF5_HEADER))
        fd.close()

    if header.decode("unicode_escape") == HDF5_HEADER:
        f, top_level = get_top_level_group(filename)
        try:
            if VERSION in list(f.keys()):
                if run_name is not None:
                    top_level = top_level[run_name]
                else:
                    # Assume that the user wants the last one
                    last_key = sorted(top_level.keys())[-1]
                    top_level = top_level[last_key]
            m = Measurements(
                filename=dest_file, copy=top_level, image_numbers=image_numbers
            )
            return m
        finally:
            f.close()
    else:
        # FIXME - add clearer exception
        raise ValueError("Received HDF5 file header was invalid")


def load_measurements_from_buffer(buf):
    from ..preferences import get_default_output_directory

    dirtgt = get_default_output_directory()
    if not (os.path.exists(dirtgt) and os.access(dirtgt, os.W_OK)):
        dirtgt = None
    fd, filename = tempfile.mkstemp(prefix="Cpmeasurements", suffix=".hdf5", dir=dirtgt)
    if sys.platform.startswith("win"):
        # Change file descriptor mode to binary
        import msvcrt

        msvcrt.setmode(fd, os.O_BINARY)
    os.write(fd, buf)
    os.close(fd)
    try:
        return load_measurements(filename)
    finally:
        os.unlink(filename)


def find_metadata_tokens(pattern):
    """Return a list of strings which are the metadata token names in a pattern

    pattern - a regexp-like pattern that specifies how to find
              metadata in a string. Each token has the form:
              "(?<METADATA_TAG>...match-exp...)" (matlab-style) or
              "\g<METADATA_TAG>" (Python-style replace)
              "(?P<METADATA_TAG>...match-exp..)" (Python-style search)
    """
    result = []
    while True:
        m = re.search("\\(\\?[<](.+?)[>]", pattern)
        if not m:
            m = re.search("\\\\g[<](.+?)[>]", pattern)
            if not m:
                m = re.search("\\(\\?P[<](.+?)[>]", pattern)
                if not m:
                    break
        result.append(m.groups()[0])
        pattern = pattern[m.end() :]
    return result


def extract_metadata(pattern, text):
    """Return a dictionary of metadata extracted from the text

    pattern - a regexp that specifies how to find
              metadata in a string. Each token has the form:
              "\(?<METADATA_TAG>...match-exp...\)" (matlab-style) or
              "\(?P<METADATA_TAG>...match-exp...\)" (Python-style)
    text - text to be searched

    We do a little fixup in here to change Matlab searches to Python ones
    before executing.
    """
    # Convert Matlab to Python
    orig_pattern = pattern
    pattern = re.sub("(\\(\\?)([<].+?[>])", "\\1P\\2", pattern)
    match = re.search(pattern, text)
    if match:
        return match.groupdict()
    else:
        raise ValueError(
            "Metadata extraction failed: regexp '%s' does not match '%s'"
            % (orig_pattern, text)
        )


def is_well_row_token(x):
    """True if the string represents a well row metadata tag"""
    return x.lower() in ("wellrow", "well_row", "row")


def is_well_column_token(x):
    """true if the string represents a well column metadata tag"""
    return x.lower() in (
        "wellcol",
        "well_col",
        "wellcolumn",
        "well_column",
        "column",
        "col",
    )


def get_agg_measurement_name(agg, object_name, feature):
    """Return the name of an aggregate measurement

    agg - one of the names in AGG_NAMES, like AGG_MEAN
    object_name - the name of the object that we're aggregating
    feature - the name of the object's measurement
    """
    return "%s_%s_%s" % (agg, object_name, feature)


def agg_ignore_feature(feature_name):
    """Return True if the feature is one to be ignored when aggregating"""
    if feature_name.startswith("Description_"):
        return True
    if feature_name.startswith("ModuleError_"):
        return True
    if feature_name.startswith("TimeElapsed_"):
        return True
    if feature_name == "Number_Object_Number":
        return True
    return False



import collections
import itertools
import logging

import numpy as np
import scyjava

from ..utilities.image import is_file_url
from ..utilities.java import jimport

from ..constants.image import MD_SIZE_S, MD_SIZE_C, MD_SIZE_Z, MD_SIZE_T, MD_SIZE_Y, MD_SIZE_X, \
    BIOFORMATS_IMAGE_EXTENSIONS

from ..reader import Reader


LOGGER = logging.getLogger(__name__)

# bioformats returns 2 for these, imageio reader returns 3
SUPPORTED_EXTENSIONS = {'.tiff', '.tif', '.ome.tif', '.ome.tiff'}
SEMI_SUPPORTED_EXTENSIONS = BIOFORMATS_IMAGE_EXTENSIONS
#TODO: disabled until CellProfiler/CellProfiler#4684 is resolved
# SUPPORTED_SCHEMES = {'file', 'http', 'https', 'ftp', 'ftps', 'omero', 's3'}
SUPPORTED_SCHEMES = {'file', 'http', 'https', 'ftp', 'ftps', 's3'}

class BioformatsReader(Reader):
    """
    Reads a variety of image formats using the bio-formats library.

    This reader is Java-based.
    """

    reader_name = "Bio-Formats"
    variable_revision_number = 1
    supported_filetypes = BIOFORMATS_IMAGE_EXTENSIONS
    supported_schemes = SUPPORTED_SCHEMES

    def __init__(self, image_file):
        self._reader = None
        self._is_file_open = False
        super().__init__(image_file)

    def get_reader(self):
        if self._reader is None:
            ImageReader = jimport("loci.formats.ImageReader")
            self._reader = ImageReader()
            self._is_file_open = False
            scyjava.when_jvm_stops(lambda: self._reader.close() if self._reader is not None else None)

        return self._reader

    def _ensure_file_open(self):
        if not self._is_file_open:
            self.get_reader().setId(self.file.path)
            self._is_file_open = True

    def read(self,
             wants_metadata_rescale=False,
             series=None,
             index=None,
             c=None,
             z=0,
             t=0,
             xywh=None,
             channel_names=None,
             ):
        """Read a single plane from the image file.
        :param wants_metadata_rescale: if `True`, return a tuple of image and a
               tuple of (min, max) for range values of image dtype gathered from
               file metadata; if `False`, returns only the image
        :param c: read from this channel. `None` = read color image if multichannel
            or interleaved RGB.
        :param z: z-stack index
        :param t: time index
        :param series: series for ``.flex`` and similar multi-stack formats
        :param index: if `None`, fall back to ``zct``, otherwise load the indexed frame
        :param xywh: a (x, y, w, h) tuple
        :param channel_names: provide the channel names for the OME metadata
        """
        self._ensure_file_open()

        FormatTools = jimport("loci.formats.FormatTools")
        ChannelSeparator = jimport("loci.formats.ChannelSeparator")
        if series is not None:
            self._reader.setSeries(series)

        if xywh is not None:
            assert isinstance(xywh, tuple) and len(xywh) == 4, "Invalid xywh tuple"
            openBytes_func = lambda x: self._reader.openBytes(x, xywh[0], xywh[1], xywh[2], xywh[3])
            width, height = xywh[2], xywh[3]
        else:
            openBytes_func = self._reader.openBytes
            width, height = self._reader.getSizeX(), self._reader.getSizeY()

        # FIXME instead of np.frombuffer use scyjava.to_python, ideally that wraps memory
        pixel_type = self._reader.getPixelType()
        little_endian = self._reader.isLittleEndian()
        if pixel_type == FormatTools.INT8:
            dtype = np.int8
        elif pixel_type == FormatTools.UINT8:
            dtype = np.uint8
        elif pixel_type == FormatTools.UINT16:
            dtype = '<u2' if little_endian else '>u2'
        elif pixel_type == FormatTools.INT16:
            dtype = '<i2' if little_endian else '>i2'
        elif pixel_type == FormatTools.UINT32:
            dtype = '<u4' if little_endian else '>u4'
        elif pixel_type == FormatTools.INT32:
            dtype = '<i4' if little_endian else '>i4'
        elif pixel_type == FormatTools.FLOAT:
            dtype = '<f4' if little_endian else '>f4'
        elif pixel_type == FormatTools.DOUBLE:
            dtype = '<f8' if little_endian else '>f8'
        if index is not None:
            image = np.frombuffer(openBytes_func(index), dtype)
            if len(image) / height / width in (3,4):
                n_channels = int(len(image) / height / width)
                if self._reader.isInterleaved():
                    image.shape = (height, width, n_channels)
                else:
                    image.shape = (n_channels, height, width)
                    image = image.transpose(1, 2, 0)
            else:
                image.shape = (height, width)
        elif self._reader.isRGB() and self._reader.isInterleaved():
            index = self._reader.getIndex(z,0,t)
            image = np.frombuffer(openBytes_func(index), dtype)
            image.shape = (height, width, self._reader.getSizeC())
            if image.shape[2] > 3:
                image = image[:, :, :3]
        elif c is not None and self._reader.getRGBChannelCount() == 1:
            index = self._reader.getIndex(z,c,t)
            image = np.frombuffer(openBytes_func(index), dtype)
            image.shape = (height, width)
        elif self._reader.getRGBChannelCount() > 1:
            n_planes = self._reader.getRGBChannelCount()
            rdr = ChannelSeparator(self._reader)
            planes = [
                np.frombuffer(
                    (rdr.openBytes(rdr.getIndex(z,i,t)) if xywh is None else
                      rdr.openBytes(rdr.getIndex(z,i,t), xywh[0], xywh[1], xywh[2], xywh[3])),
                      dtype
                ) for i in range(n_planes)]

            if len(planes) > 3:
                planes = planes[:3]
            elif len(planes) < 3:
                # > 1 and < 3 means must be 2
                # see issue #775
                planes.append(np.zeros(planes[0].shape, planes[0].dtype))
            image = np.dstack(planes)
            image.shape=(height, width, 3)
        elif self._reader.getSizeC() > 1:
            images = [
                np.frombuffer(openBytes_func(self._reader.getIndex(z,i,t)), dtype)
                      for i in range(self._reader.getSizeC())]
            image = np.dstack(images)
            image.shape = (height, width, self._reader.getSizeC())
            if not channel_names is None:
                metadata = self._reader.getMetadataStore()
                for i in range(self._reader.getSizeC()):
                    index = self._reader.getIndex(z, 0, t)
                    channel_name = metadata.getChannelName(index, i)
                    if channel_name is None:
                        channel_name = metadata.getChannelID(index, i)
                    channel_names.append(channel_name)
        elif self._reader.isIndexed():
            # TODO can we use ChannelFiller, which automatically expands channels if they
            # are "true" color and does nothing if they are "false" (e.g. applied color table)?
            #
            # The image data is indexes into a color lookup-table
            # But sometimes the table is the identity table and just generates
            # a monochrome RGB image
            #
            index = self._reader.getIndex(z,0,t)
            image = np.frombuffer(openBytes_func(index),dtype)
            lut = None
            if pixel_type in (FormatTools.INT16, FormatTools.UINT16):
                lut = self._reader.get16BitLookupTable()
            else:
                lut = self._reader.get8BitLookupTable()
            if lut is not None:
                lut = np.array([d for d in lut]).transpose()
            image.shape = (height, width)
            if (lut is not None) \
               and not np.all(lut == np.arange(lut.shape[0])[:, np.newaxis]):
                image = lut[image, :]
        else:
            index = self._reader.getIndex(z,0,t)
            image = np.frombuffer(openBytes_func(index),dtype)
            image.shape = (height,width)

        if wants_metadata_rescale == True:
            # tiff-specific
            max_sample_value = self._reader.getMetadataValue("MaxSampleValue")
            scale = None
            if max_sample_value is not None:
                try:
                    scale = scyjava.to_python(max_sample_value)
                except:
                    LOGGER.warning("WARNING: failed to get MaxSampleValue for image. Intensities may be improperly scaled.")
                    scale = None
            return image, (0.0, float(scale)) if scale else None
        return image

    def read_volume(self,
                    wants_metadata_rescale=False,
                    series=None,
                    c=None,
                    z=None,
                    t=None,
                    xywh=None,
                    channel_names=None,
                    ):
        # Whether a volume has planes stored in the z or t axis is often ambiguous.
        # If z-size > 1 we'll use z, else we'll use t. Otherwise user should choose
        # an axis to split in Images.
        bf_reader = self.get_reader()
        self._ensure_file_open()
        if series is None:
            series = 0
        bf_reader.setSeries(series)
        z_len = 0
        if z is None:
            z_len = bf_reader.getSizeZ()
            z_range = range(z_len)
        else:
            z_range = [z]
        if t is None:
            if z_len > 1:
                t_range = range(1)
            else:
                t_range = range(bf_reader.getSizeT())
        else:
            t_range = [t]

        image_stack = []
        # potentially needed for metadata rescale value
        scale = None
        for z_index, t_index in itertools.product(z_range, t_range):
            data = self.read(
                wants_metadata_rescale=wants_metadata_rescale,
                series=series,
                c=c,
                z=z_index,
                t=t_index,
                xywh=xywh,
                channel_names=channel_names,
            )
            if wants_metadata_rescale == True and type(data) is tuple:
                scale = data[1]
                data = data[0]
            image_stack.append(data)
        image_stack = np.stack(image_stack)

        if wants_metadata_rescale == True:
            return image_stack, scale
        return image_stack

    @classmethod
    def supports_format(cls, image_file, allow_open=False, volume=False):
        """This function needs to evaluate whether a given ImageFile object
        can be read by this reader class.

        Return value should be an integer representing suitability:
        -1 - 'I can't read this at all'
        1 - 'I am the one true reader for this format, don't even bother checking any others'
        2 - 'I am well-suited to this format'
        3 - 'I can read this format, but I might not be the best',
        4 - 'I can give it a go, if you must'

        The allow_open parameter dictates whether the reader is permitted to read the file when
        making this decision. If False the decision should be made using file extension only.
        Any opened files should be closed before returning.

        The volume parameter specifies whether the reader will need to return a 3D array.
        ."""
        if image_file.scheme not in SUPPORTED_SCHEMES:
            return -1
        if image_file.scheme == 'omero':
            return 1
        if image_file.full_extension in SUPPORTED_EXTENSIONS:
            return 2
        if not allow_open:
            if image_file.file_extension in SEMI_SUPPORTED_EXTENSIONS:
                return 3
            return -1

    def close(self):
        # If your reader opens a file, this needs to release any active lock,
        if self._reader is not None and scyjava.jvm_started():
            self._reader.close()
            self._reader = None

    def get_series_metadata(self):
        """Should return a dictionary with the following keys:
        Key names are in cellprofiler_core.constants.image
        MD_SIZE_S - int reflecting the number of series
        MD_SIZE_X - list of X dimension sizes, one element per series.
        MD_SIZE_Y - list of Y dimension sizes, one element per series.
        MD_SIZE_Z - list of Z dimension sizes, one element per series.
        MD_SIZE_C - list of C dimension sizes, one element per series.
        MD_SIZE_T - list of T dimension sizes, one element per series.
        MD_SERIES_NAME - list of series names, one element per series.
        """
        meta_dict = collections.defaultdict(list)
        self._ensure_file_open()
        reader = self.get_reader()
        series_count = reader.getSeriesCount()
        meta_dict[MD_SIZE_S] = series_count
        for i in range(series_count):
            reader.setSeries(i)
            meta_dict[MD_SIZE_C].append(reader.getSizeC())
            meta_dict[MD_SIZE_Z].append(reader.getSizeZ())
            meta_dict[MD_SIZE_T].append(reader.getSizeT())
            meta_dict[MD_SIZE_Y].append(reader.getSizeY())
            meta_dict[MD_SIZE_X].append(reader.getSizeX())
        return meta_dict



from ._image_subscriber import ImageSubscriber


class OutlineImageSubscriber(ImageSubscriber):
    """A setting that subscribes to the list of available object outline names
    """

    def __init__(
        self,
        text,
        value="None",
        can_be_blank=False,
        blank_text="Leave blank",
        *args,
        **kwargs,
    ):
        super(OutlineImageSubscriber, self).__init__(
            text, value, can_be_blank, blank_text, *args, **kwargs
        )

    def matches(self, setting):
        """Only match OutlineNameProvider variables"""
        return isinstance(setting, OutlineImageSubscriber)



from ._do_something import DoSomething
from ._image_set_display import ImageSetDisplay
from ._path_list_extract_button import PathListExtractButton
from ._path_list_refresh_button import PathListRefreshButton
from ._remove_setting_button import RemoveSettingButton



import io
import logging
import sys
import time
import traceback

import zmq

from ._pipeline_event_listener import PipelineEventListener
from ..analysis.reply import ImageSetSuccess, ServerExited
from ..analysis.reply import ImageSetSuccessWithDictionary
from ..analysis.reply import NoWork
from ..analysis.request import AnalysisCancel, Display, DisplayPostGroup, OmeroLogin
from ..analysis.request import DebugComplete
from ..analysis.request import DebugWaiting
from ..analysis.request import ExceptionReport
from ..analysis.request import InitialMeasurements
from ..analysis.request import Interaction
from ..analysis.request import MeasurementsReport
from ..analysis.request import PipelinePreferences
from ..analysis.request import SharedDictionary
from ..analysis.request import Work
from ..constants.worker import ED_STOP
from ..constants.worker import NOTIFY_STOP
from ..constants.worker import all_measurements
from ..measurement import Measurements
from ..utilities.measurement import load_measurements_from_buffer
from ..utilities.zmq import PollTimeoutException
from ..pipeline import CancelledException
from ..preferences import get_awt_headless
from ..preferences import set_preferences_from_dict
from ..utilities.zmq.communicable.reply.upstream_exit import UpstreamExit
from ..workspace import Workspace
from ..reader import fill_readers


LOGGER = logging.getLogger(__name__)

class Worker:
    """An analysis worker processing work at a given address

    """

    def __init__(self, context: zmq.Context, analysis_id, work_request_address, keepalive_address, with_stop_run_loop=True):

        self.context = context
        self.work_request_address = work_request_address
        self.keepalive_address = keepalive_address
        self.cancelled = False
        self.with_stop_run_loop = with_stop_run_loop
        self.current_analysis_id = analysis_id
        self.pipeline = None
        self.preferences = None
        self.initial_measurements = None

        #TODO: disabled until CellProfiler/CellProfiler#4684 is resolved
        # from ..bioformats.formatreader import set_omero_login_hook
        # set_omero_login_hook(self.omero_login_handler)

    def __enter__(self):
        # pipeline listener object
        self.pipeline_listener = PipelineEventListener(self.handle_exception)

        # Setup the work server socket
        self.work_socket = self.context.socket(zmq.REQ)
        self.work_socket.set_hwm(2000)
        self.work_socket.connect(self.work_request_address)

        # Establish a connection to the keepalive socket
        self.keepalive_socket = self.context.socket(zmq.SUB)
        # Only listen for STOP events
        self.keepalive_socket.setsockopt(zmq.SUBSCRIBE, b"STOP")
        self.keepalive_socket.connect(self.keepalive_address)

        return self

    def __exit__(self, type, value, traceback):
        if self.initial_measurements is not None:
            self.initial_measurements.close()
        self.initial_measurements = None
        self.keepalive_socket.close()
        self.work_socket.close()

    class AnalysisWorkerThreadObject(object):
        """Provide the scope needed by the analysis worker thread

        """

        def __init__(self, worker):
            self.worker = worker

        def __enter__(self):
            self.worker.enter_thread()

        def __exit__(self, type, value, tb):
            if type is not None:
                traceback.print_exception(type, value, tb)
            self.worker.exit_thread()

    def enter_thread(self):
        LOGGER.debug("Entering worker thread")

    def exit_thread(self):
        from cellprofiler_core.constants.reader import ALL_READERS
        for reader in ALL_READERS.values():
            reader.clear_cached_readers()

    def run(self):
        from cellprofiler_core.pipeline.event import CancelledException

        with self.AnalysisWorkerThreadObject(self):
            while not self.cancelled:
                try:
                    LOGGER.debug("Requesting a job")
                    # fetch a job
                    the_request = Work(self.current_analysis_id)
                    job = self.send(the_request)

                    if isinstance(job, NoWork):
                        time.sleep(0.25)  # avoid hammering server
                        # no work, currently.
                        continue
                    self.do_job(job)
                except CancelledException:
                    break

    def do_job(self, job):
        """Handle a work request to its completion

        job - request.Work
        """
        import cellprofiler_core.pipeline as cpp

        job_measurements = []
        try:
            send_dictionary = job.wants_dictionary

            LOGGER.info("Starting job")
            # Fetch the pipeline and preferences for this analysis if we don't have it
            current_pipeline = self.pipeline
            current_preferences = self.preferences
            if not current_pipeline:
                LOGGER.debug("Fetching pipeline and preferences")
                rep = self.send(PipelinePreferences(self.current_analysis_id))
                LOGGER.debug("Received pipeline and preferences response")
                preferences_dict = rep.preferences
                # update preferences to match remote values
                set_preferences_from_dict(preferences_dict)
                LOGGER.debug("Preferences updated")

                fill_readers(check_config=True)

                LOGGER.debug("Loading pipeline")

                current_pipeline = cpp.Pipeline()
                pipeline_chunks = rep.pipeline_blob.tolist()
                pipeline_io = io.StringIO("".join(pipeline_chunks))
                current_pipeline.loadtxt(pipeline_io, raise_on_error=True)

                LOGGER.debug("Pipeline loaded")
                current_pipeline.add_listener(self.pipeline_listener.handle_event)
                current_preferences = rep.preferences
                self.pipeline = current_pipeline
                self.pipeline.calculate_last_image_uses()
                self.preferences = current_preferences
            else:
                # update preferences to match remote values
                set_preferences_from_dict(current_preferences)

            # Reset the listener's state
            self.pipeline_listener.reset()
            LOGGER.debug("Getting initial measurements")
            # Fetch the path to the intial measurements if needed.

            if self.initial_measurements is None:
                LOGGER.debug("Sending initial measurements request")
                rep = self.send(InitialMeasurements(self.current_analysis_id))
                LOGGER.debug("Got initial measurements")
                self.initial_measurements = load_measurements_from_buffer(rep.buf)
            else:
                LOGGER.debug("Has initial measurements")
            # Make a copy of the measurements for writing during this job
            current_measurements = Measurements(copy=self.initial_measurements)
            all_measurements.add(current_measurements)
            job_measurements.append(current_measurements)

            successful_image_set_numbers = []
            image_set_numbers = job.image_set_numbers
            worker_runs_post_group = job.worker_runs_post_group
            LOGGER.info("Doing job: " + ",".join(map(str, image_set_numbers)))

            self.pipeline_listener.image_set_number = image_set_numbers[0]

            if not worker_runs_post_group:
                # Get the shared state from the first imageset in this run.
                shared_dicts = self.send(
                    SharedDictionary(self.current_analysis_id)
                ).dictionaries
                assert len(shared_dicts) == len(current_pipeline.modules())
                for module, new_dict in zip(current_pipeline.modules(), shared_dicts):
                    module.set_dictionary_for_worker(new_dict)

            # Run prepare group if this is the first image in the group.  We do
            # this here (even if there's no grouping in the pipeline) to ensure
            # that any changes to the modules' shared state dictionaries get
            # propagated correctly.
            should_process = True
            if current_measurements["Image", "Group_Index", image_set_numbers[0]] == 1:
                workspace = Workspace(
                    current_pipeline, None, None, None, current_measurements, None, None
                )
                if not current_pipeline.prepare_group(
                    workspace,
                    current_measurements.get_grouping_keys(),
                    image_set_numbers,
                ):
                    # exception handled elsewhere, possibly cancelling this run.
                    should_process = False
                del workspace

            # process the images
            if should_process:
                abort = False
                for image_set_number in image_set_numbers:
                    try:
                        self.pipeline_listener.image_set_number = image_set_number
                        last_workspace = current_pipeline.run_image_set(
                            current_measurements,
                            image_set_number,
                            self.interaction_handler,
                            self.display_handler,
                            self.cancel_handler,
                        )
                        if self.pipeline_listener.should_abort:
                            abort = True
                            break
                        elif self.pipeline_listener.should_skip:
                            # Report skipped image sets as successful so that
                            # analysis can complete.
                            # Report their measurements because some modules
                            # may have provided measurements before skipping.
                            pass
                        successful_image_set_numbers.append(image_set_number)
                        # Send an indication that the image set finished successfully.
                        if send_dictionary:
                            # The jobserver would like a copy of our modules'
                            # run_state dictionaries.
                            dicts = [
                                m.get_dictionary_for_worker()
                                for m in current_pipeline.modules()
                            ]
                            req = ImageSetSuccessWithDictionary(
                                self.current_analysis_id,
                                image_set_number=image_set_number,
                                shared_dicts=dicts,
                            )
                        else:
                            req = ImageSetSuccess(
                                self.current_analysis_id,
                                image_set_number=image_set_number,
                            )
                        rep = self.send(req)
                    except CancelledException:
                        LOGGER.info("Aborting job after cancellation")
                        abort = True
                    except Exception as e:
                        try:
                            LOGGER.error("Error in pipeline", exc_info=True)
                            if (
                                self.handle_exception(image_set_number=image_set_number)
                                == ED_STOP
                            ):
                                abort = True
                                break
                        except:
                            LOGGER.error(
                                "Error in handling of pipeline exception", exc_info=True
                            )
                            # this is bad.  We can't handle nested exceptions
                            # remotely so we just fail on this run.
                            abort = True

                if abort:
                    current_measurements.close()
                    job_measurements.remove(current_measurements)
                    return

                if worker_runs_post_group:
                    if not last_workspace is None:
                        last_workspace.interaction_handler = self.interaction_handler
                        last_workspace.cancel_handler = self.cancel_handler
                        last_workspace.post_group_display_handler = (
                            self.post_group_display_handler
                        )
                        # There might be an exception in this call, but it will be
                        # handled elsewhere, and there's nothing we can do for it
                        # here.
                        current_pipeline.post_group(
                            last_workspace, current_measurements.get_grouping_keys()
                        )
                        del last_workspace
                    else:
                        LOGGER.error("No workspace from last image set, cannot run post group")

            # send measurements back to server
            req = MeasurementsReport(
                self.current_analysis_id,
                buf=current_measurements.file_contents(),
                image_set_numbers=image_set_numbers,
            )
            while True:
                try:
                    rep = self.send(req, timeout=4000)
                    break
                except PollTimeoutException:
                    LOGGER.debug(f"Worker timeout on sending MeasurementsReport; reconstructing socket and retry for job {str(job.image_set_numbers)}")
                    self.work_socket.close(linger=0)
                    self.work_socket = self.context.socket(zmq.REQ)
                    self.work_socket.set_hwm(2000)
                    self.work_socket.connect(self.work_request_address)
                    continue

        except CancelledException:
            # Main thread received shutdown signal
            raise

        except Exception:
            LOGGER.error("Error in worker", exc_info=True)
            if self.handle_exception() == ED_STOP:
                raise CancelledException("Cancelling after user-requested stop")
        finally:
            # Clean up any measurements owned by us
            for m in job_measurements:
                m.close()

    def interaction_handler(self, module, *args, **kwargs):
        """handle interaction requests by passing them to the jobserver and wait for the reply."""
        # we write args and kwargs into the InteractionRequest to allow
        # more complex data to be sent by the underlying zmq machinery.
        arg_kwarg_dict = dict(
            [("arg_%d" % idx, v) for idx, v in enumerate(args)]
            + [("kwarg_%s" % name, v) for (name, v) in list(kwargs.items())]
        )
        req = Interaction(
            self.current_analysis_id,
            module_num=module.module_num,
            num_args=len(args),
            kwargs_names=list(kwargs.keys()),
            **arg_kwarg_dict,
        )
        rep = self.send(req)
        return rep.result

    def cancel_handler(self):
        """Handle a cancel request by sending AnalysisCancelRequest

        """
        self.send(AnalysisCancel(self.current_analysis_id))

    def display_handler(self, module, display_data, image_set_number):
        """handle display requests"""
        req = Display(
            self.current_analysis_id,
            module_num=module.module_num,
            display_data_dict=display_data.__dict__,
            image_set_number=image_set_number,
        )
        rep = self.send(req)

    def post_group_display_handler(self, module, display_data, image_set_number):
        req = DisplayPostGroup(
            self.current_analysis_id,
            module.module_num,
            display_data.__dict__,
            image_set_number,
        )
        rep = self.send(req)

    #TODO: disabled until CellProfiler/CellProfiler#4684 is resolved
    # def omero_login_handler(self):
    #     """Handle requests for an Omero login"""
    #     from cellprofiler_core.bioformats.formatreader import use_omero_credentials

    #     req = OmeroLogin(self.current_analysis_id)
    #     rep = self.send(req)
    #     use_omero_credentials(rep.credentials)

    def send(self, req, work_socket=None, timeout=None):
        """Send a request and receive a reply

        req - request to send

        socket - socket to use for send. Default is current work socket

        returns a reply on success. If cancelled, throws a CancelledException
        """
        if self.current_analysis_id is None:
            from cellprofiler_core.pipeline.event import CancelledException

            raise CancelledException("Can't send after cancelling")
        if work_socket is None:
            work_socket = self.work_socket
        poller = zmq.Poller()
        poller.register(self.keepalive_socket, zmq.POLLIN)
        poller.register(work_socket, zmq.POLLIN)
        req.send_only(work_socket)
        response = None
        while response is None:
            # we timeout here because we have noticed in cases of piplines with
            # many image sets (e.g. 240) and many workers (e.g. 15),
            # that for some small number of image sets (e.g 5), Workers will
            # successfully send MeasurementReports, but the Boundary.spin
            # thread never receives them. It is difficult to diagnose,
            # and difficult to fix "properly". Instead we notice we haven't
            # received an ACK for a while, and assume we're hung, and the
            # calling function of send is welcome to retry.
            # See https://github.com/CellProfiler/CellProfiler/pull/4934
            # for more details.
            poll_res = poller.poll(timeout)
            if len(poll_res) == 0:
                LOGGER.debug("Worker timeout on polling for response")
                raise PollTimeoutException
            for socket, state in poll_res:
                if state != zmq.POLLIN:
                    continue
                elif socket == self.keepalive_socket:
                    notify_msg = self.keepalive_socket.recv()
                    if notify_msg == NOTIFY_STOP:
                        LOGGER.debug("Worker received cancel notification")
                        self.cancelled = True
                        self.raise_cancel(
                            "Received stop notification while waiting for "
                            "response from %s" % str(req)
                        )
                    else:
                        LOGGER.error("Unexpected message on keepalive: " + notify_msg.decode())
                elif socket == work_socket:
                    response = req.recv(work_socket)
        if isinstance(response, (UpstreamExit, ServerExited)):
            self.raise_cancel(
                "Received UpstreamExit for analysis %s during request %s"
                % (self.current_analysis_id, str(req))
            )
        return response

    def raise_cancel(self, msg="Cancelling analysis"):
        """Handle the cleanup after some proximate cause of cancellation

        msg - reason for cancellation

        This should only be called upon detection of a server-driven
        cancellation of analysis: either UpstreamExit or a stop notification
        from the deadman thread.
        """
        from cellprofiler_core.pipeline.event import CancelledException

        LOGGER.debug(msg)
        self.cancelled = True
        if self.initial_measurements is not None:
            self.initial_measurements.close()
        self.initial_measurements = None
        self.pipeline = None
        self.preferences = None
        self.current_analysis_id = None
        raise CancelledException(msg)

    def handle_exception(self, image_set_number=None, module_name=None, exc_info=None):
        """report and handle an exception, possibly by remote debugging, returning
        how to proceed (skip or abort).

        A new socket is created for each exception report, to allow us to sidestep
        any REP/REQ state in the worker.
        """
        if self.current_analysis_id is None:
            # Analysis has been cancelled - don't initiate server interactions
            return ED_STOP
        if exc_info is None:
            t, exc, tb = sys.exc_info()
        else:
            t, exc, tb = exc_info
        filename, line_number, _, _ = traceback.extract_tb(tb, 1)[0]
        report_socket = self.context.socket(zmq.REQ)
        try:
            report_socket.connect(self.work_request_address)
        except:
            return ED_STOP  # nothing to do but give up
        try:
            req = ExceptionReport(
                self.current_analysis_id,
                image_set_number,
                module_name,
                exc_type=t.__name__,
                exc_message=str(exc),
                exc_traceback="".join(traceback.format_exception(t, exc, tb)),
                filename=filename,
                line_number=line_number,
            )
            reply = self.send(req, report_socket)
            while True:
                if reply.disposition == "DEBUG":
                    #
                    # Send DebugWaiting after we know the port #
                    #
                    debug_reply = [None]

                    def pc(port):
                        print("GOT PORT ", port)
                        debug_reply[0] = self.send(
                            DebugWaiting(self.current_analysis_id, port), report_socket,
                        )

                    print("HASH", reply.verification_hash)

                    # We get a new reply at the end, which might be "DEBUG" again.
                    reply = self.send(
                        DebugComplete(self.current_analysis_id), report_socket
                    )
                else:
                    return reply.disposition
        finally:
            report_socket.close()



from ._dependency import Dependency


class ObjectDependency(Dependency):
    """A dependency on an object labeling"""

    def __init__(
        self,
        source_module,
        destination_module,
        object_name,
        source_setting=None,
        destination_setting=None,
    ):
        super(type(self), self).__init__(
            source_module, destination_module, source_setting, destination_setting
        )
        self.__object_name = object_name

    @property
    def object_name(self):
        """The name of the objects produced by the source and used by the dest"""
        return self.__object_name

    def __str__(self):
        return "Object: %s" % self.object_name



builtin_modules = {
    "align": "Align",
    "groups": "Groups",
    "images": "Images",
    "loaddata": "LoadData",
    "metadata": "Metadata",
    "namesandtypes": "NamesAndTypes",
}
all_modules: dict = {}
svn_revisions: dict = {}
pymodules: list = []
badmodules: list = []
do_not_override = ["set_settings", "create_from_handles", "test_valid", "module_class"]
should_override = ["create_settings", "settings", "run"]
renamed_modules = {
    "ApplyThreshold": "Threshold",
    "CalculateImageOverlap": "MeasureImageOverlap",
    "CropObjects": "SaveCroppedObjects",
    "Dilation": "DilateImage",
    "MeasureNeurons": "MeasureObjectSkeleton",
    "MeasureCorrelation": "MeasureColocalization",
    "NoiseReduction": "ReduceNoise",
    "ReassignObjectNumbers": "SplitOrMergeObjects",
    "SplitIntoContiguousObjects": "SplitOrMergeObjects",
    "Erosion": "ErodeImage",
}
replaced_modules = {
    "LoadImageDirectory": ["LoadData"],
    "GroupMovieFrames": ["LoadData"],
    "IdentifyPrimLoG": ["IdentifyPrimaryObjects"],
    "FileNameMetadata": ["LoadData"],
    "LoadSingleImage": ["LoadData"],
    "LoadImages": ["LoadData"],
}
depricated_modules = ["CorrectIllumination_Calculate_kate", "SubtractBackground"]
unimplemented_modules = ["LabelImages", "Restart", "SplitOrSpliceMovie"]
